import unittest
import os
import shutil

from aix360.algorithms.contrastive import CEMExplainer
from aix360.algorithms.contrastive import KerasClassifier
from aix360.datasets import MNISTDataset
from keras.models import model_from_json

import matplotlib.pyplot as plt
import numpy as np


class TestCEMExplainer(unittest.TestCase):

    def load_model(self, model_json_file, model_wt_file):

        # read model json file
        with open(model_json_file, 'r') as f:
            model = model_from_json(f.read())

        # read model weights file
        model.load_weights(model_wt_file)

        return model


    def test_CEM(self):

        # set up

        _PATH = os.path.dirname(os.path.realpath(__file__))

        fp = os.path.join(_PATH, "cem_tests_results")
        if not os.path.exists(fp):
            os.makedirs(fp)

        data = MNISTDataset()

        ae_js = os.path.join(_PATH, "../../aix360/models/CEM/mnist_AE_1_decoder.json")
        ae_wt = os.path.join(_PATH, "../../aix360/models/CEM/mnist_AE_1_decoder.h5")

        ae_model = self.load_model(ae_js, ae_wt)

        model_js = os.path.join(_PATH, "../../aix360/models/CEM/mnist.json")
        model_wt = os.path.join(_PATH, "../../aix360/models/CEM/mnist")

        mnist_model = self.load_model(model_js, model_wt)

        mymodel = KerasClassifier(mnist_model)

        explainer = CEMExplainer(mymodel)

        # Explain an input instance
        image_id = 340
        input_image = data.test_data[image_id]
        plt.imshow(input_image[:,:,0], cmap="gray")
        plt.savefig(os.path.join(fp, 'input_image_340_d3.png'))

        # check prediction of model is "3" as expected
        self.assertEqual(mymodel.predict_classes(np.expand_dims(input_image, axis=0)), 3)

        # Obtain Pertinent Negative explanation
        arg_max_iter = 1000
        arg_b = 9
        arg_init_const = 10.0
        arg_mode = "PN"
        arg_kappa = 10
        arg_beta = 1e-1
        arg_gamma = 100

        (adv_pn, delta_pn, _) = explainer.explain_instance(np.expand_dims(input_image, axis=0),
                                                           arg_mode, ae_model, arg_kappa, arg_b,
                                                           arg_max_iter, arg_init_const, arg_beta, arg_gamma)

        # Obtain Pertinent Positive explanation
        arg_mode = "PP"

        (adv_pp, delta_pp, _) = explainer.explain_instance(np.expand_dims(input_image, axis=0),
                                                           arg_mode, ae_model, arg_kappa, arg_b,
                                                           arg_max_iter, arg_init_const, arg_beta, arg_gamma)

        fig0 = (input_image[:, :, 0] + 0.5) * 255
        fig1 = (adv_pn[0, :, :, 0] + 0.5) * 255
        fig2 = (fig1 - fig0)  # rescaled delta_pn
        fig3 = (adv_pp[0, :, :, 0] + 0.5) * 255
        fig4 = (fig0 - fig3)  # rescaled delta_pp

        _, axarr = plt.subplots(1, 5)
        axarr[0].set_title("Orig" + "(" + str(mymodel.predict_classes(np.expand_dims(input_image, axis=0))[0]) + ")")
        axarr[1].set_title("Orig + PN" + "(" + str(mymodel.predict_classes(adv_pn)[0]) + ")")
        axarr[2].set_title("PN")
        axarr[3].set_title("Orig + PP")
        axarr[4].set_title("PP" + "(" + str(mymodel.predict_classes(delta_pp)[0]) + ")")

        axarr[0].imshow(fig0, cmap="gray")
        axarr[1].imshow(fig1, cmap="gray")
        axarr[2].imshow(fig2, cmap="gray")
        axarr[3].imshow(fig3, cmap="gray")
        axarr[4].imshow(fig4, cmap="gray")
        plt.savefig(os.path.join(fp, 'predict_classes_4.png'))

        if os.path.exists(fp):
            shutil.rmtree(fp)

if __name__ == '__main__':
    unittest.main()


import unittest

from aix360.algorithms.dipvae import DIPVAEExplainer

import torch
import torchvision.datasets as dset
import torchvision.transforms as transforms
import argparse
import numpy as np


class FakeFMnistDataset():
    def __init__(self, batch_size=256, subset_size=256, test_batch_size=256):
        trans = transforms.Compose([transforms.ToTensor()])

        root = './data_fake_fmnist'
        train_set = dset.FakeData(image_size=(1, 28, 28),transform=transforms.ToTensor())
        test_set = dset.FakeData(image_size=(1, 28, 28),transform=transforms.ToTensor())

        indices = torch.randperm(len(train_set))[:subset_size]
        train_set = torch.utils.data.Subset(train_set, indices)

        self.train_loader = torch.utils.data.DataLoader(
            dataset=train_set,
            batch_size=batch_size,
            shuffle=True)
        self.test_loader = torch.utils.data.DataLoader(
            dataset=test_set,
            batch_size=test_batch_size,
            shuffle=False)

        self.name = "fakemnist"
        self.data_dims = [28, 28, 1]
        self.train_size = len(self.train_loader)
        self.test_size = len(self.test_loader)
        self.range = [0.0, 1.0]
        self.batch_size = batch_size
        self.num_training_instances = len(train_set)
        self.num_test_instances = len(test_set)
        self.likelihood_type = 'gaussian'
        self.output_activation_type = 'sigmoid'

    def next_batch(self):
        for x, y in self.train_loader:
            #print(x.size())
            #x = np.reshape(x, (-1, 28, 28, 1))
            yield x, None

    def next_test_batch(self):
        for x, y in self.test_loader:
            #x = np.reshape(x, (-1, 28, 28, 1))
            yield x, None


class TestDIPVAEExplainer(unittest.TestCase):

    def test_DIPVAEExplainer(self):

        # Load the dataset object
        dataset_obj = FakeFMnistDataset()
        self.assertIsNotNone(dataset_obj)

        cuda_available = torch.cuda.is_available()

        # Initialize model arguments
        dipvaeii_args = argparse.Namespace()

        dipvaeii_args.model = 'dipvae-ii'

        dipvaeii_args.activation_type = 'relu'
        dipvaeii_args.num_nodes = 1200
        dipvaeii_args.latent_dim = 10
        dipvaeii_args.num_gen_layers = 3
        dipvaeii_args.num_inference_layers = 2

        dipvaeii_args.step_size = 0.001
        dipvaeii_args.lambda_diag_factor = 10.0
        dipvaeii_args.lambda_offdiag = 0.001

        dipvaeii_args.seed = 0

        # Run for 2 epochs for testing elbo increase
        dipvaeii_args.num_epochs = 2

        # Initialize the explainer model
        dipvaeii_explainer = DIPVAEExplainer(model_args=dipvaeii_args, dataset=dataset_obj, cuda_available=cuda_available)

        # Fit the generative model.
        loss_epoch_list = dipvaeii_explainer.fit()

        # Testing elbo increase.
        self.assertGreaterEqual(loss_epoch_list[1], loss_epoch_list[0])

        for x, _ in dataset_obj.next_test_batch():
            input_images = x.squeeze().view(-1, np.prod(dataset_obj.data_dims))
            break

        convert_and_reshape = lambda x: x.cpu().data.numpy().reshape([-1] + dataset_obj.data_dims)
        edited_images = dipvaeii_explainer.explain(input_images=input_images,
                                                              edit_dim_id=0,
                                                              edit_dim_value=-1.0, edit_z_sample=True)

        # Make sure edited images are different
        self.assertGreater(np.linalg.norm(
            np.abs(convert_and_reshape(input_images) - convert_and_reshape(edited_images)))
            , 0.0)


if __name__ == '__main__':
    unittest.main()




# Copyright 2019 IBM Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.




# COPYRIGHT

# All contributions by François Chollet:
# Copyright (c) 2015 - 2019, François Chollet.
# All rights reserved.

# All contributions by Google:
# Copyright (c) 2015 - 2019, Google, Inc.
# All rights reserved.

# All contributions by Microsoft:
# Copyright (c) 2017 - 2019, Microsoft, Inc.
# All rights reserved.

# All other contributions:
# Copyright (c) 2015 - 2019, the respective contributors.
# All rights reserved.

# Each contributor holds copyright over their respective contributions.
# The project versioning (Git) records all such contribution source information.

# LICENSE

# The MIT License (MIT)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.




"""
#Trains a ResNet on the CIFAR10 dataset.
ResNet v1:
[Deep Residual Learning for Image Recognition
](https://arxiv.org/pdf/1512.03385.pdf)
ResNet v2:
[Identity Mappings in Deep Residual Networks
](https://arxiv.org/pdf/1603.05027.pdf)
Model|n|200-epoch accuracy|Original paper accuracy |sec/epoch GTX1080Ti
:------------|--:|-------:|-----------------------:|---:
ResNet20   v1|  3| 92.16 %|                 91.25 %|35
ResNet32   v1|  5| 92.46 %|                 92.49 %|50
ResNet44   v1|  7| 92.50 %|                 92.83 %|70
ResNet56   v1|  9| 92.71 %|                 93.03 %|90
ResNet110  v1| 18| 92.65 %|            93.39+-.16 %|165
ResNet164  v1| 27|     - %|                 94.07 %|  -
ResNet1001 v1|N/A|     - %|                 92.39 %|  -
&nbsp;
Model|n|200-epoch accuracy|Original paper accuracy |sec/epoch GTX1080Ti
:------------|--:|-------:|-----------------------:|---:
ResNet20   v2|  2|     - %|                     - %|---
ResNet32   v2|N/A| NA    %|            NA         %| NA
ResNet44   v2|N/A| NA    %|            NA         %| NA
ResNet56   v2|  6| 93.01 %|            NA         %|100
ResNet110  v2| 12| 93.15 %|            93.63      %|180
ResNet164  v2| 18|     - %|            94.54      %|  -
ResNet1001 v2|111|     - %|            95.08+-.14 %|  -
"""

from __future__ import print_function
import keras
from keras.layers import Dense, Conv2D, BatchNormalization, Activation
from keras.layers import AveragePooling2D, Input, Flatten
from keras.optimizers import Adam
from keras.callbacks import ModelCheckpoint, LearningRateScheduler
from keras.callbacks import ReduceLROnPlateau
from keras.preprocessing.image import ImageDataGenerator
from keras.regularizers import l2
from keras import backend as K
from keras.models import Model
from keras.datasets import cifar10
import numpy as np
import os
import json
from collections import namedtuple


# A named tuple that specified various hyper parameters that are generic for any keras model and in particular needed for 
# resnet models defined here.
HParams = namedtuple('HParams','complexity_param,num_classes,batch_size,epochs,checkpoint_path,lr_reducer,lr_scheduler,optimizer')


def lr_schedule(epoch):
    """Learning Rate Schedule
    Learning rate is scheduled to be reduced after 80, 120, 160, 180 epochs.
    Called automatically every epoch as part of callbacks during training.
    # Arguments
        epoch (int): The number of epochs
    # Returns
        lr (float32): learning rate
    """
    lr = 1e-3
    if epoch > 180:
        lr *= 0.5e-3
    elif epoch > 160:
        lr *= 1e-3
    elif epoch > 120:
        lr *= 1e-2
    elif epoch > 80:
        lr *= 1e-1
    print('Learning rate: ', lr)
    return lr


def resnet_layer(inputs,
                 num_filters=16,
                 kernel_size=3,
                 strides=1,
                 activation='relu',
                 batch_normalization=True,
                 conv_first=True):
    """2D Convolution-Batch Normalization-Activation stack builder
    # Arguments
        inputs (tensor): input tensor from input image or previous layer
        num_filters (int): Conv2D number of filters
        kernel_size (int): Conv2D square kernel dimensions
        strides (int): Conv2D square stride dimensions
        activation (string): activation name
        batch_normalization (bool): whether to include batch normalization
        conv_first (bool): conv-bn-activation (True) or
            bn-activation-conv (False)
    # Returns
        x (tensor): tensor as input to the next layer
    """
    conv = Conv2D(num_filters,
                  kernel_size=kernel_size,
                  strides=strides,
                  padding='same',
                  kernel_initializer='he_normal',
                  kernel_regularizer=l2(1e-4))

    x = inputs
    if conv_first:
        x = conv(x)
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
    else:
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
        x = conv(x)
    return x


def resnet_v1(input_shape, num_res_blocks, num_classes=10):
    """ResNet Version 1 Model builder [a]
    Stacks of 2 x (3 x 3) Conv2D-BN-ReLU
    Last ReLU is after the shortcut connection.
    At the beginning of each stage, the feature map size is halved (downsampled)
    by a convolutional layer with strides=2, while the number of filters is
    doubled. Within each stage, the layers have the same number filters and the
    same number of filters.
    Features maps sizes:
    stage 0: 32x32, 16
    stage 1: 16x16, 32
    stage 2:  8x8,  64
    The Number of parameters is approx the same as Table 6 of [a]:
    ResNet20 0.27M
    ResNet32 0.46M
    ResNet44 0.66M
    ResNet56 0.85M
    ResNet110 1.7M
    # Arguments
        input_shape (tensor): shape of input image tensor
        depth (int): number of core convolutional layers
        num_res_blocks (int): number of res blocks to be used. 
        num_classes (int): number of classes (CIFAR10 has 10)
    # Returns
        model (Model): Keras model instance
    """
    #if (depth - 2) % 6 != 0:
    #    raise ValueError('depth should be 6n+2 (eg 20, 32, 44 in [a])')
    # Start model definition.
    num_filters = 16
    #num_res_blocks = int((depth - 2) / 6)

    inputs = Input(shape=input_shape)
    x = resnet_layer(inputs=inputs)
    # Instantiate the stack of residual units
    for stack in range(3):
        for res_block in range(num_res_blocks):
            strides = 1
            if stack > 0 and res_block == 0:  # first layer but not first stack
                strides = 2  # downsample
            y = resnet_layer(inputs=x,
                             num_filters=num_filters,
                             strides=strides)
            y = resnet_layer(inputs=y,
                             num_filters=num_filters,
                             activation=None)
            if stack > 0 and res_block == 0:  # first layer but not first stack
                # linear projection residual shortcut connection to match
                # changed dims
                x = resnet_layer(inputs=x,
                                 num_filters=num_filters,
                                 kernel_size=1,
                                 strides=strides,
                                 activation=None,
                                 batch_normalization=False)
            x = keras.layers.add([x, y])
            x = Activation('relu')(x)
        num_filters *= 2

    # Add classifier on top.
    # v1 does not use BN after last shortcut connection-ReLU
    x = AveragePooling2D(pool_size=8)(x)
    y = Flatten()(x)
    outputs = Dense(num_classes,
                    activation='softmax',
                    kernel_initializer='he_normal')(y)

    # Instantiate model.
    model = Model(inputs=inputs, outputs=outputs)
    return model





# Model parameter
# ----------------------------------------------------------------------------
#           |      | 200-epoch | Orig Paper| 200-epoch | Orig Paper| sec/epoch
# Model     |  n   | ResNet v1 | ResNet v1 | ResNet v2 | ResNet v2 | GTX1080Ti
#           |v1(v2)| %Accuracy | %Accuracy | %Accuracy | %Accuracy | v1 (v2)
# ----------------------------------------------------------------------------
# ResNet20  | 3 (2)| 92.16     | 91.25     | -----     | -----     | 35 (---)
# ResNet32  | 5(NA)| 92.46     | 92.49     | NA        | NA        | 50 ( NA)
# ResNet44  | 7(NA)| 92.50     | 92.83     | NA        | NA        | 70 ( NA)
# ResNet56  | 9 (6)| 92.71     | 93.03     | 93.01     | NA        | 90 (100)
# ResNet110 |18(12)| 92.65     | 93.39+-.16| 93.15     | 93.63     | 165(180)
# ResNet164 |27(18)| -----     | 94.07     | -----     | 94.54     | ---(---)
# ResNet1001| (111)| -----     | 92.39     | -----     | 95.08+-.14| ---(---)
# ---------------------------------------------------------------------------







import pandas as pd
import unittest

from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score

from aix360.algorithms.rbm import FeatureBinarizer, BRCGExplainer, BooleanRuleCG


class TestBooleanmRuleCG(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.bc = load_breast_cancer()

    def test_classification(self):
        bc_df = pd.DataFrame(self.bc.data, columns=self.bc.feature_names)
        X_train, X_test, Y_train, Y_test = train_test_split(bc_df, self.bc.target, test_size = 0.2, random_state = 31)
        fb = FeatureBinarizer(negations=True)
        X_train_fb = fb.fit_transform(X_train)
        X_test_fb = fb.transform(X_test)

        self.assertEqual(len(X_train_fb.columns), 540)
        self.assertEqual(len(X_test_fb.columns), 540)

        boolean_model = BooleanRuleCG(silent=True)
        explainer = BRCGExplainer(boolean_model)
        explainer.fit(X_train_fb, Y_train)
        Y_pred = explainer.predict(X_test_fb)

        self.assertGreater(accuracy_score(Y_test, Y_pred), 0.9)
        self.assertGreater(precision_score(Y_test, Y_pred), 0.9)
        self.assertGreater(recall_score(Y_test, Y_pred), 0.9)
        self.assertGreater(f1_score(Y_test, Y_pred), 0.9)

        explanation = explainer.explain()
        self.assertEqual(explanation['rules'], [
          'compactness error > 0.01 AND worst concavity <= 0.22 AND worst symmetry <= 0.28',
          'mean texture <= 15.46 AND mean concavity <= 0.15 AND area error <= 54.16',
          'fractal dimension error > 0.00 AND worst area <= 680.60 AND worst concave points <= 0.18',
          'mean concave points <= 0.05 AND perimeter error <= 3.80 AND worst area <= 930.88 AND worst smoothness <= 0.16'
        ])


if __name__ == '__main__':
    unittest.main()



import os

import pandas as pd


def default_preprocessing(df):
    df = df[df['RACEV2X'].isin(['1','2'])]
    df = df[df['HISPANX'] == 2]

    racedict = {1.0: 0.0,2.0: 1.0}
    df = df.assign(RACEV2X = df['RACEV2X'].replace(to_replace = racedict))
    df = df.rename(columns = {'RACEV2X' : 'RACE3'})

    genderdict = {1.0: 0.0,2.0: 1.0}
    df = df.assign(SEX = df['SEX'].replace(to_replace = genderdict))
    df = df.rename(columns = {'SEX' : 'GENDER'})
    df = df.rename(columns = {'PERWT15F' : 'PERSONWT'})
    df = df.rename(columns = {'REGION31' : 'REGION'})
    df = df.rename(columns = {'TTLP15X' : 'INCOME_M'})
    df = df.rename(columns = {'TOTEXP15' : 'HEALTHEXP'})

    # #df = df.drop('HISPANX',axis=1)  .column dropping can be taken care of by simply not including in features_to_keep
    # df = df[df['PANEL'] == 19]
    # #df = df.drop('PANEL',axis=1)
    # lessE = df['TOTEXP15'] <= 5000.0
    # df.loc[lessE,'TOTEXP15'] = 0.0
    # moreE = df['TOTEXP15'] > 5000.0
    # df.loc[moreE,'TOTEXP15'] = 1.0

    df = df[df['REGION'] >= 0] # remove values -1
    df = df[df['AGE31X'] >= 0] # remove values -1

    df = df[df['MARRY31X'] >= 0] # remove values -1, -7, -8, -9
    df = df[df['INCOME_M'] >= 0]

    df = df[(df[['EDRECODE','FTSTU31X','ACTDTY31','HONRDC31','RTHLTH31','MNHLTH31','HIBPDX','CHDDX','ANGIDX',
                 'MIDX','OHRTDX','STRKDX','EMPHDX','CHBRON31','CHOLDX','CANCERDX','DIABDX',
                 'JTPAIN31','ARTHDX','ARTHTYPE','ASTHDX','ADHDADDX','PREGNT31','WLKLIM31',
                 'ACTLIM31','SOCLIM31','COGLIM31','DFHEAR42','DFSEE42','ADSMOK42',
                 'PHQ242','EMPST31','POVCAT15','INSCOV15']] >= -1).all(1)]  #for all other categorical features, remove values < -1
    df = df[['PANEL', 'REGION','AGE31X','GENDER','RACE3','MARRY31X',     # dropped 'EDUCYR', 'HIDEG' as data distributions are weird across panels,
             'EDRECODE','FTSTU31X','ACTDTY31','HONRDC31','RTHLTH31','MNHLTH31','HIBPDX','CHDDX','ANGIDX',
             'MIDX','OHRTDX','STRKDX','EMPHDX','CHBRON31','CHOLDX','CANCERDX','DIABDX',
             'JTPAIN31','ARTHDX','ARTHTYPE','ASTHDX','ADHDADDX','PREGNT31','WLKLIM31',
             'ACTLIM31','SOCLIM31','COGLIM31','DFHEAR42','DFSEE42','ADSMOK42','PCS42',
             'MCS42','K6SUM42','PHQ242','EMPST31','POVCAT15','INSCOV15','INCOME_M','HEALTHEXP','PERSONWT']]

    df['HEALTHEXP'].mean(), df['INCOME_M'].mean()
    return df

class MEPSDataset():
    """
    The Medical Expenditure Panel Survey (MEPS) [#]_ data consists of large scale surveys of families and individuals, 
    medical providers, and employers, and collects data on health services used, costs & frequency of services,
    demographics, health status and conditions, etc., of the respondents.

    This specific dataset contains MEPS survey data for calendar year 2015 obtained in rounds 3, 4, and 5 of Panel 19,
    and rounds 1, 2, and 3 of Panel 20.
    See :file:`aix360/data/meps_data/README.md` for more details on the dataset and instructions on downloading/processing the data.

    References:
        .. [#] `Medical Expenditure Panel Survey data <https://meps.ahrq.gov/mepsweb/>`_

    """

    def __init__(self, custom_preprocessing=default_preprocessing, dirpath=None):
        self._dirpath = dirpath
        if not self._dirpath:
            self._dirpath = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                '..', 'data','meps_data')

        self._filepath = os.path.join(self._dirpath, 'h181.csv')
        try:
            df = pd.read_csv(self._filepath, sep=',', na_values=[])
        except IOError as err:
            print("IOError: {}".format(err))
            print("To use this class, please place the heloc_dataset.csv:")
            print("file, as-is, in the folder:")
            print("\n\t{}\n".format(os.path.abspath(os.path.join(
               os.path.abspath(__file__), '..', 'data','meps_data'))))
            import sys
            sys.exit(1)

        if custom_preprocessing:
            self._data = custom_preprocessing(df)

    def data(self):
        return self._data



import os
import numpy as np
import pandas as pd
import cvxpy as cvx
from sklearn.metrics import confusion_matrix, accuracy_score

from .beam_search import beam_search, beam_search_K1


class BooleanRuleCG(object):
    """BooleanRuleCG is a directly interpretable supervised learning method
    for binary classification that learns a Boolean rule in disjunctive
    normal form (DNF) or conjunctive normal form (CNF) using column generation (CG).
    AIX360 implements a heuristic beam search version of BRCG that is less 
    computationally intensive than the published integer programming version [#NeurIPS2018]_.

    References:
        .. [#NeurIPS2018] `S. Dash, O. Günlük, D. Wei, "Boolean decision rules via
           column generation." Neural Information Processing Systems (NeurIPS), 2018.
           <https://papers.nips.cc/paper/7716-boolean-decision-rules-via-column-generation.pdf>`_
    """
    def __init__(self,
        lambda0=0.001,
        lambda1=0.001,
        CNF=False,
        iterMax=100,
        K=10,
        D=10,
        B=5,
        eps=1e-6,
        solver='ECOS',
        verbose=False,
        silent=False):
        """
        Args:
            lambda0 (float, optional): Complexity - fixed cost of each clause
            lambda1 (float, optional): Complexity - additional cost for each literal
            CNF (bool, optional): CNF instead of DNF
            iterMax (int, optional): Column generation - maximum number of iterations
            K (int, optional): Column generation - maximum number of columns generated per iteration
            D (int, optional): Column generation - maximum degree
            B (int, optional): Column generation - beam search width
            eps (float, optional): Numerical tolerance on comparisons
            solver (str, optional): Linear programming - solver
            verbose (bool, optional): Linear programming - verboseness
            silent (bool, optional): Silence overall algorithm messages
        """
        # Complexity parameters
        self.lambda0 = lambda0      # fixed cost of each clause
        self.lambda1 = lambda1      # additional cost per literal
        # CNF instead of DNF
        self.CNF = CNF
        # Column generation parameters
        self.iterMax = iterMax      # maximum number of iterations
        self.K = K                  # maximum number of columns generated per iteration
        self.D = D                  # maximum degree
        self.B = B                  # beam search width
        # Numerical tolerance on comparisons
        self.eps = eps
        # Linear programming parameters
        self.solver = solver        # solver
        self.verbose = verbose      # verboseness
        # Silence output
        self.silent = silent

    def fit(self, X, y):
        """Fit model to training data.

        Args:
            X (DataFrame): Binarized features with MultiIndex column labels
            y (array): Binary-valued target variable
        Returns:
            BooleanRuleCG: Self
        """
        if not self.silent:
            print('Learning {} rule with complexity parameters lambda0={}, lambda1={}'\
                  .format('CNF' if self.CNF else 'DNF', self.lambda0, self.lambda1))
        if self.CNF:
            # Flip labels for CNF
            y = 1 - y
        # Positive (y = 1) and negative (y = 0) samples
        P = np.where(y > 0.5)[0]
        Z = np.where(y < 0.5)[0]
        nP = len(P)
        n = len(y)

        # Initialize with empty and singleton conjunctions, i.e. X plus all-ones feature
        # Feature indicator and conjunction matrices
        z = pd.DataFrame(np.eye(X.shape[1], X.shape[1]+1, 1, dtype=int), index=X.columns)
        A = np.hstack((np.ones((X.shape[0],1), dtype=int), X))
        # Iteration counter
        self.it = 0

        # Formulate master LP
        # Variables
        w = cvx.Variable(A.shape[1], nonneg=True)
        xi = cvx.Variable(nP, nonneg=True)
        # Objective function (no penalty on empty conjunction)
        cs = self.lambda0 + self.lambda1 * z.sum().values
        cs[0] = 0
        obj = cvx.Minimize(cvx.sum(xi) / n + cvx.sum(A[Z,:] * w) / n + cs * w)
        # Constraints
        constraints = [xi + A[P,:] * w >= 1]

        # Solve problem
        prob = cvx.Problem(obj, constraints)
        prob.solve(solver=self.solver, verbose=self.verbose)
        if not self.silent:
            print('Initial LP solved')

        # Extract dual variables
        r = np.ones_like(y, dtype=float) / n
        r[P] = -constraints[0].dual_value

        # Beam search for conjunctions with negative reduced cost
        # Most negative reduced cost among current variables
        UB = np.dot(r, A) + cs
        UB = min(UB.min(), 0)
        v, zNew, Anew = beam_search(r, X, self.lambda0, self.lambda1,
                                    K=self.K, UB=UB, D=self.D, B=self.B, eps=self.eps)

        while (v < -self.eps).any() and (self.it < self.iterMax):
            # Negative reduced costs found
            self.it += 1
            if not self.silent:
                print('Iteration: {}, Objective: {:.4f}'.format(self.it, prob.value))

            # Add to existing conjunctions
            z = pd.concat([z, zNew], axis=1, ignore_index=True)
            A = np.concatenate((A, Anew), axis=1)

            # Reformulate master LP
            # Variables
            w = cvx.Variable(A.shape[1], nonneg=True)
            # Objective function
            cs = np.concatenate((cs, self.lambda0 + self.lambda1 * zNew.sum().values))
            obj = cvx.Minimize(cvx.sum(xi) / n + cvx.sum(A[Z,:] * w) / n + cs * w)
            # Constraints
            constraints = [xi + A[P,:] * w >= 1]

            # Solve problem
            prob = cvx.Problem(obj, constraints)
            prob.solve(solver=self.solver, verbose=self.verbose)

            # Extract dual variables
            r[P] = -constraints[0].dual_value

            # Beam search for conjunctions with negative reduced cost
            # Most negative reduced cost among current variables
            UB = np.dot(r, A) + cs
            UB = min(UB.min(), 0)
            v, zNew, Anew = beam_search(r, X, self.lambda0, self.lambda1,
                                        K=self.K, UB=UB, D=self.D, B=self.B, eps=self.eps)

        # Save generated conjunctions and LP solution
        self.z = z
        self.wLP = w.value

        r = np.full(nP, 1./n)
        self.w = beam_search_K1(r, pd.DataFrame(1-A[P,:]), 0, A[Z,:].sum(axis=0) / n + cs,
                                UB=r.sum(), D=100, B=2*self.B, eps=self.eps, stopEarly=False)[1].values.ravel()
        if len(self.w) == 0:
            self.w = np.zeros_like(self.wLP, dtype=int)

    def compute_conjunctions(self, X):
        """Compute conjunctions of features as specified in self.z.
        
        Args:
            X (DataFrame): Binarized features with MultiIndex column labels
        Returns:
            array: A -- Conjunction values
        """
        try:
            A = 1 - (np.dot(1 - X, self.z) > 0) # Changed matmul to dot, because failed on some machines
        except AttributeError:
            print("Attribute 'z' does not exist, please fit model first.")
        return A

    def predict(self, X):
        """Predict class labels.

        Args:
            X (DataFrame): Binarized features with MultiIndex column labels
        Returns:
            array: y -- Predicted labels
        """
        # Compute conjunctions of features
        A = self.compute_conjunctions(X)
        # Predict labels
        if self.CNF:
            # Flip labels since model is actually a DNF for Y=0
            return 1 - (np.dot(A, self.w) > 0)
        else:
            return (np.dot(A, self.w) > 0).astype(int)

    def explain(self, maxConj=None, prec=2):
        """Return rules comprising the model.

        Args:
            maxConj (int, optional): Maximum number of conjunctions to show
            prec (int, optional): Number of decimal places to show for floating-value thresholds
        Returns:
            Dictionary containing
            
            * isCNF (bool): flag signaling whether model is CNF or DNF
            * rules (list): selected conjunctions formatted as strings
        """
        # Selected conjunctions
        z = self.z.loc[:, self.w > 0.5]
        truncate = (maxConj is not None) and (z.shape[1] > maxConj)
        nConj = maxConj if truncate else z.shape[1]

        """
        if self.CNF:
            print('Predict Y=0 if ANY of the following rules are satisfied, otherwise Y=1:')
        else:
            print('Predict Y=1 if ANY of the following rules are satisfied, otherwise Y=0:')
        """

        # Sort conjunctions by increasing order
        idxSort = z.sum().sort_values().index[:nConj]
        # Iterate over sorted conjunctions
        conj = []
        for i in idxSort:
            # MultiIndex of features participating in rule i
            idxFeat = z.index[z[i] > 0]
            # String representations of features
            strFeat = idxFeat.get_level_values(0) + ' ' + idxFeat.get_level_values(1)\
                + ' ' + idxFeat.get_level_values(2).to_series()\
                .apply(lambda x: ('{:.' + str(prec) + 'f}').format(x) if type(x) is float else str(x))
            # String representation of rule
            strFeat = strFeat.str.cat(sep=' AND ')
            conj.append(strFeat)

        return {
            'isCNF': self.CNF,
            'rules': conj
        }



# Directly interpretable supervised explainers

import abc
import sys

# Ensure compatibility with Python 2/3
if sys.version_info >= (3, 4):
    ABC = abc.ABC
else:
    ABC = abc.ABCMeta(str('ABC'), (), {})


class DISExplainer(ABC):
    """
    DISExplainer is the base class for Directly Interpretable Supervised Explainers (DISE).
    Such explainers attempt to train explainable models directly from data.
    Examples include Boolean Decision Rules via Column Generation[#1]_, Generalized Linear Rule Models[#2]_,
    TED: Teaching AI to Explain its Decisions[#3]_, etc.

    References:
        .. [#1] Boolean Decision Rules via Column Generation, NIPS 2018.
        Dash, Gunluk, Wei. https://arxiv.org/abs/1805.09901
        .. [#2] Generalized Linear Rule Models, ICML 2019.
        Dennis Wei, Sanjeeb Dash, Tian Gao, Oktay Gunluk.
        https://arxiv.org/abs/1906.01761
        .. [#3] TED: Teaching AI to Explain its Decisions, AIES 2019.
        Michael Hind, Dennis Wei, Murray Campbell, Noel C. F. Codella, Amit Dhurandhar,
        Aleksandra Mojsilovic, Karthikeyan Natesan Ramamurthy, Kush R. Varshney.
        https://arxiv.org/abs/1811.04896
    """

    def __init__(self, *argv, **kwargs):
        """
        Initialize a DISExplainer object.
        ToDo: check common steps that need to be distilled here.
        """

    @abc.abstractmethod
    def set_params(self, *argv, **kwargs):
        """
        Set parameters for the explainer.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def explain(self, *argv, **kwargs):
        """
        Explain the model
        """
        raise NotImplementedError

    @abc.abstractmethod
    def fit(self, *argv, **kwargs):
        """
        Fit a model on data.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def predict(self, *argv, **kwargs):
        """
        Prediction for a batch of inputs (numpy array)
        """
        raise NotImplementedError



import numpy as np
import pandas as pd
from sklearn.base import TransformerMixin
from sklearn.preprocessing import OneHotEncoder, StandardScaler

class FeatureBinarizer(TransformerMixin):
    '''Transformer for binarizing categorical and ordinal (including continuous) features.
    
    For use with LogisticRuleRegression and LinearRuleRegression
    '''
    def __init__(self, colCateg=[], numThresh=9, negations=False, threshStr=False, returnOrd=False, **kwargs):
        """
        Args:
            colCateg (list): Categorical features ('object' dtype automatically treated as categorical)
            numThresh (int): Number of quantile thresholds used to binarize ordinal variables
            negations (bool): Append negations
            threshStr (bool): Convert thresholds on ordinal features to strings
            returnOrd (bool): Also return standardized ordinal features
        """
        # List of categorical columns
        if type(colCateg) is pd.Series:
            self.colCateg = colCateg.tolist()
        elif type(colCateg) is not list:
            self.colCateg = [colCateg]
        else:
            self.colCateg = colCateg
        # Number of quantile thresholds used to binarize ordinal features
        self.numThresh = numThresh
        self.thresh = {}
        # whether to append negations
        self.negations = negations
        # whether to convert thresholds on ordinal features to strings
        self.threshStr = threshStr
        # Also return standardized ordinal features
        self.returnOrd = returnOrd

    def fit(self, X):
        '''Fit FeatureBinarizer to data
        
        Args:
            X (DataFrame): Original features
        Returns:
            FeatureBinarizer: Self
            self.maps (dict): Mappings for unary/binary columns
            self.enc (dict): OneHotEncoders for categorical columns
            self.thresh (dict(array)): Thresholds for ordinal columns
            self.NaN (list): Ordinal columns containing NaN values
            self.ordinal (list): Ordinal columns
            self.scaler (StandardScaler): StandardScaler for ordinal columns
        '''
        data = X
        # Quantile probabilities
        quantProb = np.linspace(1. / (self.numThresh + 1.), self.numThresh / (self.numThresh + 1.), self.numThresh)
        # Initialize
        maps = {}
        enc = {}
        thresh = {}
        NaN = []
        if self.returnOrd:
            ordinal = []

        # Iterate over columns
        for c in data:
            # number of unique values
            valUniq = data[c].nunique()

            # Constant or binary column
            if valUniq <= 2:
                # Mapping to 0, 1
                maps[c] = pd.Series(range(valUniq), index=np.sort(data[c].unique()))

            # Categorical column
            elif (c in self.colCateg) or (data[c].dtype == 'object'):
                # OneHotEncoder object
                enc[c] = OneHotEncoder(sparse=False, dtype=int, handle_unknown='ignore')
                # Fit to observed categories
                enc[c].fit(data[[c]])

            # Ordinal column
            elif np.issubdtype(data[c].dtype, np.dtype(int).type) \
                | np.issubdtype(data[c].dtype, np.dtype(float).type):
                # Few unique values
                if valUniq <= self.numThresh + 1:
                    # Thresholds are sorted unique values excluding maximum
                    thresh[c] = np.sort(data[c].unique())[:-1]
                # Many unique values
                else:
                    # Thresholds are quantiles excluding repetitions
                    thresh[c] = data[c].quantile(q=quantProb).unique()
                if data[c].isnull().any():
                    # Contains NaN values
                    NaN.append(c)
                if self.returnOrd:
                    ordinal.append(c)

            else:
                print(("Skipping column '" + str(c) + "': data type cannot be handled"))
                continue

        self.maps = maps
        self.enc = enc
        self.thresh = thresh
        self.NaN = NaN
        if self.returnOrd:
            self.ordinal = ordinal
            # Fit StandardScaler to ordinal features
            self.scaler = StandardScaler().fit(data[ordinal])
        return self

    def transform(self, X):
        '''Binarize features
        
        Args:
            X (DataFrame): Original features
        Returns:
            A (DataFrame): Binarized features with MultiIndex column labels
            Xstd (DataFrame, optional): Standardized ordinal features
        '''
        data = X
        maps = self.maps
        enc = self.enc
        thresh = self.thresh
        NaN = self.NaN

        # Initialize dataframe
        A = pd.DataFrame(index=data.index,
                         columns=pd.MultiIndex.from_arrays([[], [], []], names=['feature', 'operation', 'value']))

        # Iterate over columns
        for c in data:
            # Constant or binary column
            if c in maps:
                # Rename values to 0, 1
                A[(str(c), '', '')] = data[c].map(maps[c])
                if self.negations:
                    A[(str(c), 'not', '')] = 1 - A[(str(c), '', '')]

            # Categorical column
            elif c in enc:
                # Apply OneHotEncoder
                Anew = enc[c].transform(data[[c]])
                Anew = pd.DataFrame(Anew, index=data.index, columns=enc[c].categories_[0].astype(str))
                if self.negations:
                    # Append negations
                    Anew = pd.concat([Anew, 1 - Anew], axis=1, keys=[(str(c), '=='), (str(c), '!=')])
                else:
                    Anew.columns = pd.MultiIndex.from_product([[str(c)], ['=='], Anew.columns])
                # Concatenate
                A = pd.concat([A, Anew], axis=1)

            # Ordinal column
            elif c in thresh:
                # Threshold values to produce binary arrays
                Anew = (data[c].values[:, np.newaxis] <= thresh[c]).astype(int)
                if self.negations:
                    # Append negations
                    Anew = np.concatenate((Anew, 1 - Anew), axis=1)
                    ops = ['<=', '>']
                else:
                    ops = ['<=']
                # Convert to dataframe with column labels
                if self.threshStr:
                    Anew = pd.DataFrame(Anew, index=data.index,
                                        columns=pd.MultiIndex.from_product([[str(c)], ops, thresh[c].astype(str)]))
                else:
                    Anew = pd.DataFrame(Anew, index=data.index,
                                        columns=pd.MultiIndex.from_product([[str(c)], ops, thresh[c]]))
                if c in NaN:
                    # Ensure that rows corresponding to NaN values are zeroed out
                    indNull = data[c].isnull()
                    Anew.loc[indNull] = 0
                    # Add NaN indicator column
                    Anew[(str(c), '==', 'NaN')] = indNull.astype(int)
                    if self.negations:
                        Anew[(str(c), '!=', 'NaN')] = (~indNull).astype(int)
                # Concatenate
                A = pd.concat([A, Anew], axis=1)

            else:
                print(("Skipping column '" + str(c) + "': data type cannot be handled"))
                continue

        if self.returnOrd:
            # Standardize ordinal features
            Xstd = self.scaler.transform(data[self.ordinal])
            Xstd = pd.DataFrame(Xstd, index=data.index, columns=self.ordinal)
            # Fill NaN with mean (which is now zero)
            Xstd.fillna(0, inplace=True)
            return A, Xstd
        else:
            return A   



import os
import sys
import requests
import pandas as pd
import xport, csv



def default_preprocessing(df):
    return df

class CDCDataset():
    """
    The CDC (Center for Disease Control and Prevention) questionnaire datasets [#]_ are surveys conducted
    by the organization involving 1000s of civilians about various facets of daily life. There are 44
    questionnaires that collect data about income, occupation, health, early childhood and many other
    behavioral and lifestyle aspects of people living in the US. These questionnaires are thus a rich
    source of information indicative of the quality of life of many civilians. More information about
    each questionaire and the type of answers are available in the following reference.

    References:
        .. [#] `NHANES 2013-2014 Questionnaire Data
           <https://wwwn.cdc.gov/nchs/nhanes/search/datapage.aspx?Component=Questionnaire&CycleBeginYear=2013>`_
    """

    def __init__(self, custom_preprocessing=default_preprocessing, dirpath=None):

        self._cdcfileinfo, self._cdcweb, self._cdcfiles = self._cdc_files_info()
        self._dirpath = dirpath
        if not self._dirpath:
            self._dirpath = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                '..', 'data','cdc_data')

        self._csv_path = os.path.join(self._dirpath, 'csv')
        if not os.path.exists(self._dirpath):
            os.mkdir(self._dirpath)

        for f in self._cdcfiles:
            try:
                filename = os.path.join(self._dirpath, f)
                if not os.path.exists(filename):
                    print("Downloading file {}".format(f))
                    file = requests.get(os.path.join(self._cdcweb, f), allow_redirects=True)
                    open(filename, 'wb').write(file.content)
            except IOError as err:
                print("IOError: {}".format(err))
                sys.exit(1)

        self._convert_xpt_to_csv()
        #if custom_preprocessing:
        #    self._data = custom_preprocessing(df)


    def _cdc_files_info(self):
        # List of files (i.e. questionnaires) in the CDC dataset. The following 4 files were ignored due to processing issues.
        # RXQ_RX_H: Prescription Medications
        # SMQ_H: Smoking - Cigarette Use
        # PUQMEC_H: Pesticide Use
        # RXQ_DRUG.xpt: Prescription Medications - Drug Information
        # If the errors can be fixed, they can be added to this list.

        cdcfileinfo = ['Acculturation', 'Alcohol Use', 'Blood Pressure & Cholesterol', 'Cardiovascular Health',
                    'Cognitive Functioning', 'Consumer Behavior', 'Creatine Kinase', 'Current Health Status',
                    'Dermatology', 'Diabetes', 'Diet Behavior & Nutrition', 'Disability', 'Drug Use', 'Early Childhood',
                    'Food Security', 'Health Insurance', 'Hepatitis', 'Hospital Utilization & Access to Care',
                    'Housing Characteristics', 'Immunization', 'Income', 'Kidney Conditions - Urology',
                    'Medical Conditions', 'Mental Health - Depression Screener',
                    'Occupation', 'Oral Health', 'Osteoporosis', 'Physical Activity',
                    'Physical Functioning', 'Preventive Aspirin Use','Reproductive Health',
                    'Sexual Behavior', 'Sleep Disorders', 'Smoking - Household Smokers', 'Smoking - Recent Tobacco Use',
                    'Smoking - Secondhand Smoke Exposure', 'Taste & Smell', 'Volatile Toxicant (Subsample)',
                    'Weight History', 'Weight History - Youth'
                   ]
        cdcweb = 'https://wwwn.cdc.gov/Nchs/Nhanes/2013-2014/'
        cdcfiles = ["ACQ_H.XPT",
        "ALQ_H.XPT",
        "BPQ_H.XPT",
        "CDQ_H.XPT",
        "CFQ_H.XPT",
        "CBQ_H.XPT",
        "CKQ_H.XPT",
        "HSQ_H.XPT",
        "DEQ_H.XPT",
        "DIQ_H.XPT",
        "DBQ_H.XPT",
        "DLQ_H.XPT",
        "DUQ_H.XPT",
        "ECQ_H.XPT",
        "FSQ_H.XPT",
        "HIQ_H.XPT",
        "HEQ_H.XPT",
        "HUQ_H.XPT",
        "HOQ_H.XPT",
        "IMQ_H.XPT",
        "INQ_H.XPT",
        "KIQ_U_H.XPT",
        "MCQ_H.XPT",
        "DPQ_H.XPT",
        "OCQ_H.XPT",
        "OHQ_H.XPT",
        "OSQ_H.XPT",
        "PAQ_H.XPT",
        "PFQ_H.XPT",
        "RXQASA_H.XPT",
        "RHQ_H.XPT",
        "SXQ_H.XPT",
        "SLQ_H.XPT",
        "SMQFAM_H.XPT",
        "SMQRTU_H.XPT",
        "SMQSHS_H.XPT",
        "CSQ_H.XPT",
        "VTQ_H.XPT",
        "WHQ_H.XPT",
        "WHQMEC_H.XPT"]

        return cdcfileinfo, cdcweb, cdcfiles

    def _convert_xpt_to_csv(self):
        if not os.path.exists(self._csv_path):
            os.mkdir(self._csv_path)

        for i in range(len(self._cdcfiles)):
            f = self._cdcfiles[i]
            finfo = self._cdcfileinfo[i]
            xptfile = os.path.join(self._dirpath, f)
            csvfile = os.path.join(self._csv_path, f)
            csvfile = os.path.splitext(csvfile)[0]
            csvfile = csvfile + ".csv"
            if not os.path.exists(csvfile):
                print("converting ", finfo, ": ", xptfile, " to ", csvfile)
                with open(xptfile, 'rb') as in_xpt:
                    with open(csvfile, 'w',newline='') as out_csv:
                        reader = xport.Reader(in_xpt)
                        writer = csv.writer(out_csv, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        writer.writerow(reader.fields)
                        for row in reader:
                            writer.writerow(row)


    def get_csv_file(self, filename):
        return pd.read_csv(os.path.join(self._csv_path, filename))

    def get_csv_file_names(self):
        return [os.path.splitext(x)[0]+".csv" for x in self._cdcfiles]





import time
import six
import sys

#import cifar_input
import numpy as np
import aix360.algorithms.profwt.resnet_keras_model
import tensorflow as tf
from sklearn.utils import shuffle
import json


dataset='cifar10' # 'cifar10 or cifar100.'



def fully_connected(x, out_dim):
  """FullyConnected layer
  
  Parameters:
  x (Tensor): Input tensor to the fully connected layer
  out_dim (int): Output dimension of the fully connected layer.
  
  Return: 
  The Tensor corresponding to the fully connected layer output.
  """
  w = tf.get_variable(
      'DW', [x.get_shape()[1], out_dim],
      initializer=tf.uniform_unit_scaling_initializer(factor=1.0))
  b = tf.get_variable('biases', [out_dim],
                      initializer=tf.constant_initializer())
  return tf.nn.xw_plus_b(x, w, b)

def probe_train_eval(probe_train_input,y_train1,num_classes,probe_eval_input,y_train2,to_save_probe_model_filename):
  """ 
  Given the input corresponding to a previously flattened layer output stored in a file, this trains a simple logistic model
  with the flattened layer output at the input to predict the label. We call this logistic model a probe classifier. Probe 
  classifier is then evaulated on a set of layer outputs of a different dataset.
  
  Parameters:
  # probe_train_input (numpy array): Layer output treated as the input to train the probe classifier on. Dim - (num of samples_1 x flattened layer output dimension)
  # y_train1 (numpy array): Labels corresponding to probe_train_input that is used to train the probe classifier. (num of samples_1 x num of classes)
  # probe_eval_input (numpy array): Layer output belonging to (perhaps) a different dataset to evaluate probe classifier on. Dimension - (num of samples_2 x flattened layer output dimension)
  # y_train2 (numpy array): Labels corresponding to probe_eval_input that is used to evaluate the probe classifier. Dimension - (num of samples_2 x num of classes)
  # to_save_probe_model_filename (string): Filename to save the probe classifier trained on probe_eval_input.

  Return:
  array_logits (numpy array): Matrix of logits of the Probe Classifier of dimensions (number of samples x num of classes)
  array_pred (numpy array): Matrix of probabilities of the Probe Classifier of dimensions (number of samples x num of classes)
  """

  probe_val=probe_train_input

  # Tensorflow Graph Definition
  columns=probe_val.shape[1]
  probe_optimizer='mom'

  probe_input=tf.placeholder(tf.float32,shape=(None,columns))
  probe_labels=tf.placeholder(tf.float32,shape=(None,num_classes))
  #with tf.variable_scope('probe_'+str(j)+'_first')
  #  probe_first_layer=fully_connected(probe_input,first_layer_dim)
  probe_first_layer=probe_input

  with tf.variable_scope('probe_logits'):
    probe_logits=fully_connected(probe_first_layer,num_classes)
    probe_predictions = tf.nn.softmax(probe_logits)
    probe_cost_1 = tf.nn.softmax_cross_entropy_with_logits(
          logits=probe_logits, labels=probe_labels)
    probe_cost_2 = tf.reduce_mean(probe_cost_1, name='xent')

  # Tensorflow training operations defined.
  learning_rate=0.1
  trainable_variables = tf.trainable_variables()
  grads = tf.gradients(probe_cost_2,trainable_variables)

  if probe_optimizer == 'sgd':
    optimizer = tf.train.GradientDescentOptimizer(learning_rate)
  elif probe_optimizer == 'mom':
    optimizer = tf.train.MomentumOptimizer(learning_rate,0.9)
  probe_train_op = optimizer.apply_gradients(zip(grads, trainable_variables),name='train_step')

  probe_truth = tf.argmax(probe_labels, axis=1)
  predictions = tf.argmax(probe_predictions, axis=1)
  probe_precision = tf.reduce_mean(tf.to_float(tf.equal(predictions,probe_truth)))
  epochs=200
  num_examples=len(probe_val)
  batch_size=128

  saver=tf.train.Saver()
  probe_val_2=probe_eval_input

  print("Start Training Probe Model.....")
  # Training Session that runs the training operation after initialization
  with tf.Session() as sess:
    sess.run(tf.global_variables_initializer())
    global_step=0
    for i in range(epochs):
      print(i)
      x_t, y_t = shuffle(probe_val,y_train1)
      for offset in range(0, num_examples, batch_size):
        end = offset + batch_size
        batch_x, batch_y = x_t[offset:end,:], y_t[offset:end,:]
        sess.run(probe_train_op,feed_dict={probe_input: batch_x, probe_labels: batch_y})
        global_step=global_step+1
        #print('Training Step:',global_step)
      if (i+1)%100==0:
        saver.save(sess,to_save_probe_model_filename)
    print("Starting to Evaluate Probe Model Confidences and Saving in a File.....")
    saver.restore(sess,to_save_probe_model_filename)
    batch_size=500
    num_examples=len(probe_val_2)
    array_logits=[]
    array_pred=[]
    precision=0
    l=0
    for offset in range(0, num_examples, batch_size):
      print (l)
      end = offset + batch_size
      batch_x, batch_y = probe_val_2[offset:end,:], y_train2[offset:end,:]
      if l==0:
        array_logits=sess.run(probe_logits,feed_dict={probe_input: batch_x, probe_labels : batch_y})
        array_pred=sess.run(probe_predictions,feed_dict={probe_input: batch_x, probe_labels : batch_y})
      else:
        arr_dummy_logits=sess.run(probe_logits,feed_dict={probe_input: batch_x, probe_labels: batch_y})
        arr_dummy_pred=sess.run(probe_predictions,feed_dict={probe_input: batch_x, probe_labels: batch_y})
        array_logits=np.append(np.array(array_logits),np.array(arr_dummy_logits),axis=0) 
        array_pred=np.append(np.array(array_pred),np.array(arr_dummy_pred),axis=0)
      l=l+1
      #precision=precision+sess.run(probe_precision,feed_dict={probe_input: batch_x, probe_labels: batch_y})
    #print("Precision on train2",precision/len(range(0, num_examples, batch_size)))
  print(array_logits.shape)
  print(array_pred.shape)
  #np.save(to_save_logits_filename,array_logits)
  #np.save(to_save_pred_filename,array_pred)
  print("Probe Confidences/Logits Saved...")
  tf.reset_default_graph()
  return (array_logits,array_pred)








import unittest
import tensorflow as tf
import sys, os
import numpy as np
import random

""" This file is a unit test for CEM_MAF. It
1. Loads the model to be explained
2. Loads and predicts on a specific image.
3. Creates an explainer object with two attributes
4. Creates a pertinent positive for the image.

"""

_PATH = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(_PATH, '../..'))

from aix360.algorithms.contrastive.CEM_MAF import CEM_MAFImageExplainer
from aix360.algorithms.contrastive.CEM_MAF_utils import CELEBAModel
from aix360.algorithms.contrastive.classifiers import KerasClassifier
from aix360.algorithms.contrastive.dwnld_CEM_MAF_celebA import dwnld_CEM_MAF_celebA
from aix360.datasets.celeba_dataset import CelebADataset

class TestCEM_MAFImageExplainer(unittest.TestCase):

    def test_CEM_MAFExplainer(self):

        dwnld = dwnld_CEM_MAF_celebA()

        with tf.Session() as sess: # Open TensorFlow Session
            random.seed(121)
            np.random.seed(1211)
            sess.run(tf.global_variables_initializer())
        
            # Fix attributes to use for test
            attributes = ["Brown_Hair", "High_Cheekbones"]

            # Download pretrained celebA model
            local_path_models = os.path.join(_PATH,'../../aix360/models/CEM_MAF')
            celebA_model_file = dwnld.dwnld_celebA_model(local_path_models)

            # Download attribute functions
            attr_model_files = dwnld.dwnld_celebA_attributes(local_path_models, attributes)

            # Load the pretrained celebA model
            model_file = os.path.join(_PATH,'../../aix360/models/CEM_MAF/celebA')
            loaded_model = CELEBAModel(restore=model_file, use_softmax=False).model
        
            mymodel = KerasClassifier(loaded_model)
        
            # Select an input image, download, load, and process it
            img_ids = [15]
            img_id = img_ids[0]
            local_path_img =  os.path.join(_PATH,'../../aix360/data/celeba_data')
            img_files = dwnld.dwnld_celebA_data(local_path_img, img_ids)
            dataset_obj = CelebADataset(local_path_img) # use the CelebA dataset class
            input_img = dataset_obj.get_img(img_id)
            input_img = np.clip(input_img/2, -0.5, 0.5)
            
            # Predict on input image
            _, orig_class, _ = mymodel.predict_long(input_img)
        
            # Compute classes
            young_flag = orig_class % 2
            smile_flag = (orig_class // 2) % 2
            sex_flag = (orig_class // 4) % 2
            arg_img_name = os.path.join(local_path_img, "{}_img.png".format(img_id))
            print("Image:{}, pred:{}".format(arg_img_name, orig_class))
            print("Male:{}, Smile:{}, Young:{}".format(sex_flag, smile_flag, young_flag))
                
            # designate path to aix360 - needed to find paths to attribute files
            aix360_path = os.path.join(_PATH, '../../aix360') 

            # Create explainer object with two attributes
            attributes = ["Brown_Hair", "High_Cheekbones"]
            explainer = CEM_MAFImageExplainer(mymodel, attributes, aix360_path)
            
            # Set parameters for Pertinent Positive explanation
            arg_mode = 'PP'
            arg_kappa = 5
            arg_gamma = 100.0
            arg_beta = 0.1
            arg_binary_search_steps = 1
            arg_max_iterations = 10
            arg_initial_const = 10
            
            # Run optimizer to find a pertinent positive
            (adv_pp, __, __) = explainer.explain_instance(sess, input_img, None, arg_mode, arg_kappa, 
                                arg_binary_search_steps, arg_max_iterations, 
                                arg_initial_const, arg_gamma, arg_beta)
        
            _, adv_class, _ = mymodel.predict_long(adv_pp)
        
            # Compute class of PP
            young_flag = adv_class % 2
            smile_flag = (adv_class // 2) % 2
            sex_flag = (adv_class // 4) % 2
            print("Pertinent positive pred:{}".format(adv_class))
            print("Male:{}, Smile:{}, Young:{}".format(sex_flag, smile_flag, young_flag))
            
            # remove celebA model file, all attribute model files, and all data files
            os.remove(celebA_model_file[0])
            
            for model_file in attr_model_files:
                os.remove(model_file)
                
            for data_file in img_files:
                os.remove(data_file)
            
if __name__ == '__main__':
    unittest.main()



from .CEM import CEMExplainer
from .CEM_MAF import CEM_MAFImageExplainer
from .classifiers import KerasClassifier
from .CEM_MAF_utils import load_AE, CELEBAModel
from .dwnld_CEM_MAF_celebA import dwnld_CEM_MAF_celebA



import numpy as np
import pandas as pd
from sklearn.linear_model import Lasso, Ridge

import matplotlib.pyplot as plt

from .beam_search import beam_search_K1


class LinearRuleRegression(object):
    """Linear Rule Regression is a directly interpretable supervised learning
    method that performs linear regression on rule-based features.
    """
    def __init__(self,
      lambda0=0.05,
      lambda1=0.01,
      useOrd=False,
      debias=True,
      K=1,
      iterMax=200,
      B=1,
      wLB=0.5,
      stopEarly=False, eps=1e-6):

        """
        Args:
            lambda0 (float, optional): Regularization - fixed cost of each rule
            lambda1 (float, optional): Regularization - additional cost of each literal in rule
            useOrd (bool, optional): Also use standardized numerical features
            debias (bool, optional): Re-fit final solution without regularization
            K (int, optional): Column generation - maximum number of columns generated per iteration
            iterMax (int, optional): Column generation - maximum number of iterations
            B (int, optional): Column generation - beam search width
            wLB (float, optional): Column generation - weight on lower bound in evaluating nodes
            stopEarly (bool, optional): Column generation - stop after current degree once improving column found
            eps (float, optional): Numerical tolerance on comparisons
        """
        # Use standardized ordinal features
        self.useOrd = useOrd
        # Regularization parameters
        self.lambda0 = lambda0      # fixed cost of term
        self.lambda1 = lambda1      # cost per literal
        self.debias = debias        # re-fit final solution without regularization
        # Column generation parameters
        self.K = K                  # maximum number of columns generated per iteration
        self.iterMax = iterMax      # maximum number of iterations
        self.B = B                  # beam search width
        self.wLB = wLB              # weight on lower bound in evaluating nodes
        self.stopEarly = stopEarly  # stop after current degree once improving column found
        # Numerical tolerance on comparisons
        self.eps = eps

    def fit(self, X, y, Xstd=None):
        """Fit model to training data.

        Args:
            X (DataFrame): Binarized features with MultiIndex column labels
            y (array): Target variable
            Xstd (DataFrame, optional): Standardized numerical features
        Returns:
            LinearRuleRegression: Self
        """
        # Initialization
        # Number of samples
        n = X.shape[0]
        # Initialize with X itself i.e. singleton conjunctions
        # Feature indicator and conjunction matrices
        z = pd.DataFrame(np.eye(X.shape[1], dtype=int), index=X.columns)
#        A = X.values
        # Remove negations
        indPos = X.columns.get_level_values(1).isin(['', '<=', '=='])
        z = z.loc[:,indPos]
        A = X.loc[:,indPos].values
        # Scale conjunction matrix to account for non-uniform penalties
        A = A * self.lambda0 / (self.lambda0 + self.lambda1 * z.sum().values)
        if self.useOrd:
            self.namesOrd = Xstd.columns
            numOrd = Xstd.shape[1]
            # Scale ordinal features to have similar std as "average" binary feature
            Astd = 0.4 * Xstd.values
        # Iteration counter
        self.it = 0
        # Mean absolute deviation from mean as upper bound on lambdas that lead to non-constant solutions
        self.mu = y.mean()
        MADM = np.abs(y - self.mu).mean()
        # Lasso object
        lr = Lasso(alpha=self.lambda0 * MADM / 2, selection='cyclic')
#        lr = LassoLars(alpha=self.lambda0 * MADM / 2, normalize=False)

        # Fit Lasso model
        if self.useOrd:
            B = np.concatenate((Astd, A), axis=1)
            lr.fit(B, y)
            # Initial residual
            r = (lr.predict(B) - y) / n / (MADM/2)
        else:
            lr.fit(A, y)
            # Initial residual
            r = (lr.predict(A) - y) / n / (MADM/2)
        # Most "negative" subderivative among current variables (undo scaling)
        UB = -np.abs(np.dot(r, A))
        UB *= (self.lambda0 + self.lambda1 * z.sum().values) / self.lambda0
        UB += self.lambda0 + self.lambda1 * z.sum().values
        UB = min(UB.min(), 0)

        # Beam search for conjunctions with subdifferentials that exclude zero
        vp, zp, Ap = beam_search_K1(r, X, self.lambda0, self.lambda1,
            UB=UB, B=self.B, wLB=self.wLB, eps=self.eps, stopEarly=self.stopEarly)
        vn, zn, An = beam_search_K1(-r, X, self.lambda0, self.lambda1,
            UB=UB, B=self.B, wLB=self.wLB, eps=self.eps, stopEarly=self.stopEarly)
        v = np.append(vp, vn)

        while (v < UB).any() and (self.it < self.iterMax):
            # Subdifferentials excluding zero exist, continue
            self.it += 1
            zNew = pd.concat([zp, zn], axis=1, ignore_index=True)
            Anew = np.concatenate((Ap, An), axis=1)

            # K conjunctions with largest subderivatives in absolute value
            idxLargest = np.argsort(v)[:self.K]
            v = v[idxLargest]
            zNew = zNew.iloc[:,idxLargest]
            Anew = Anew[:,idxLargest]
            # Scale new conjunction matrix to account for non-uniform penalties
            Anew = Anew * self.lambda0 / (self.lambda0 + self.lambda1 * zNew.sum().values)

            # Add to existing conjunctions
            z = pd.concat([z, zNew], axis=1, ignore_index=True)
            A = np.concatenate((A, Anew), axis=1)
            # Fit Lasso model
            if self.useOrd:
                B = np.concatenate((Astd, A), axis=1)
                lr.fit(B, y)
                # Residual
                r = (lr.predict(B) - y) / n / (MADM/2)
            else:
                lr.fit(A, y)
                # Residual
                r = (lr.predict(A) - y) / n / (MADM/2)

            # Most "negative" subderivative among current variables (undo scaling)
            UB = -np.abs(np.dot(r, A))
            UB *= (self.lambda0 + self.lambda1 * z.sum().values) / self.lambda0
            UB += self.lambda0 + self.lambda1 * z.sum().values
            UB = min(UB.min(), 0)

            # Beam search for conjunctions with subdifferentials that exclude zero
            vp, zp, Ap = beam_search_K1(r, X, self.lambda0, self.lambda1,
                UB=UB, B=self.B, wLB=self.wLB, eps=self.eps, stopEarly=self.stopEarly)
            vn, zn, An = beam_search_K1(-r, X, self.lambda0, self.lambda1,
                UB=UB, B=self.B, wLB=self.wLB, eps=self.eps, stopEarly=self.stopEarly)
            v = np.append(vp, vn)

        # Restrict model to conjunctions with nonzero coefficients
        try:
            idxNonzero = np.where(np.abs(lr.coef_) > self.eps)[0]
            if self.useOrd:
                # Nonzero indices of standardized and rule features
                self.idxNonzeroOrd = idxNonzero[idxNonzero < numOrd]
                nnzOrd = len(self.idxNonzeroOrd)
                idxNonzeroRules = idxNonzero[idxNonzero >= numOrd] - numOrd
                if self.debias and len(idxNonzero):
                    # Re-fit Lasso model with effectively no regularization
                    z = z.iloc[:,idxNonzeroRules]
#                    lr.alpha = self.eps
                    lr = Ridge(alpha=self.eps)
                    lr.fit(B[:,idxNonzero], y)
                    idxNonzero = np.where(np.abs(lr.coef_) > self.eps)[0]
                    # Nonzero indices of standardized and rule features
                    idxNonzeroOrd2 = idxNonzero[idxNonzero < nnzOrd]
                    self.idxNonzeroOrd = self.idxNonzeroOrd[idxNonzeroOrd2]
                    idxNonzeroRules = idxNonzero[idxNonzero >= nnzOrd] - nnzOrd
                self.z = z.iloc[:,idxNonzeroRules]
                lr.coef_ = lr.coef_[idxNonzero]
            else:
                if self.debias and len(idxNonzero):
                    # Re-fit Lasso model with effectively no regularization
                    z = z.iloc[:,idxNonzero]
#                    lr.alpha = self.eps
                    lr = Ridge(alpha=self.eps)
                    lr.fit(A[:,idxNonzero], y)
                    idxNonzero = np.where(np.abs(lr.coef_) > self.eps)[0]
                self.z = z.iloc[:,idxNonzero]
                lr.coef_ = lr.coef_[idxNonzero]
        except AttributeError:
            # Model has no coefficients except intercept
            self.z = z
        self.lr = lr

    def compute_conjunctions(self, X):
        """Compute conjunctions of features as specified in self.z.

        Args:
            X (DataFrame): Binarized features with MultiIndex column labels
        Returns:
            array: A -- Feature conjunction values, shape (X.shape[0], self.z.shape[1])
        """
        try:
            #A = 1 - (np.matmul(1 - X, self.z) > 0)
            A = 1 - (((1 - X) @ self.z) > 0)
        except AttributeError:
            print("Attribute 'z' does not exist, please fit model first.")
        # Scale conjunctions as done in training
        A = A * self.lambda0 / (self.lambda0 + self.lambda1 * self.z.sum().values)
        return A

    def predict(self, X, Xstd=None):
        """Predict responses.

        Args:
            X (DataFrame): Binarized features with MultiIndex column labels
            Xstd (DataFrame, optional): Standardized numerical features
        Returns:
            array: yhat -- Predicted responses
        """
        if len(self.lr.coef_):
            # Compute conjunctions of features
            A = self.compute_conjunctions(X)
            if self.useOrd:
                # Selected ordinal features scaled as in training
                Astd = 0.4 * Xstd.values[:,self.idxNonzeroOrd]
                # Predict probabilities
                B = np.concatenate((Astd, A), axis=1)
                return self.lr.predict(B)
            else:
                # Predict labels
                return self.lr.predict(A)
        else:
            return np.full(X.shape[0], self.mu)

    def explain(self, maxCoeffs=None, highDegOnly=False, prec=2):
        """Return DataFrame holding model features and their coefficients.
        
        Args:
            maxCoeffs (int, optional): Maximum number of rules/numerical features to show
            highDegOnly (bool, optional): Only show higher-degree rules
            prec (int, optional): Number of decimal places to show for floating-value thresholds
        Returns:
            DataFrame: dfExpl -- Rules/numerical features and their coefficients
        """
        # Number of ordinal features used
        if self.useOrd:
            nnzOrd = len(self.idxNonzeroOrd)
        else:
            nnzOrd = 0
        if highDegOnly:
            # Restrict to higher-degree rules
            coeffs = self.lr.coef_[nnzOrd:][self.z.sum() > 1]
            nnzOrd = 0
            z = self.z.loc[:, self.z.sum() > 1]
        else:
            coeffs = self.lr.coef_
            z = self.z
        # Initialize DataFrame to be printed
        truncate = (maxCoeffs is not None) and (len(coeffs) > maxCoeffs)
        nRows = maxCoeffs + 1 if truncate else len(coeffs) + 1
        dfExpl = pd.DataFrame(index=range(nRows), columns=['rule','coefficient'])

        # Intercept term
        dfExpl.at[0, 'rule'] = '(intercept)'
        dfExpl.at[0, 'coefficient'] = self.lr.intercept_
        # Sort coefficients from largest to smallest
        idxSort = np.abs(coeffs).argsort()[:-nRows:-1]
        dfExpl.loc[1:, 'coefficient'] = coeffs[idxSort]
        # Iterate over sorted coefficients
        for (row, i) in enumerate(idxSort):
            if i < nnzOrd:
                # Ordinal feature
                dfExpl.at[row+1, 'rule'] = self.namesOrd[self.idxNonzeroOrd[i]]
            else:
                # MultiIndex of features participating in rule i
                idxFeat = z.index[z.iloc[:,i - nnzOrd] > 0]
                # String representations of features
                strFeat = idxFeat.get_level_values(0) + ' ' + idxFeat.get_level_values(1)\
                    + ' ' + idxFeat.get_level_values(2).to_series()\
                    .apply(lambda x: ('{:.' + str(prec) + 'f}').format(x) if type(x) is float else str(x))
#                        + ' ' + idxFeat.get_level_values(2).astype(str)
                # String representation of rule
                dfExpl.at[row+1, 'rule'] = strFeat.str.cat(sep=' AND ')

        # Return dataframe holding model features and their coefficients
        return dfExpl

    def visualize(self, Xorig, fb, features=None):
        """Plot generalized additive model component, which includes first-degree rules 
        and linear functions of unbinarized ordinal features but excludes higher-degree rules.
        
        Args:
            Xorig (DataFrame): Original unbinarized features
            fb: FeatureBinarizer object used to binarize features
            features (list, optional): Subset of features to be plotted
        """
        # Number of ordinal features used
        if self.useOrd:
            nnzOrd = len(self.idxNonzeroOrd)
        else:
            nnzOrd = 0

        # Initialize terms Series and x values for plots
        terms = pd.Series(index=pd.MultiIndex.from_arrays([[],[],[]], names=self.z.index.names))
        xPlot = {}

        # Iterate over ordinal features
        for i in range(nnzOrd):
            # Restrict to specified features
            f = self.namesOrd[self.idxNonzeroOrd[i]]
            if (features is not None) and (f not in features):
                continue
            # Append term
            terms = terms.append(pd.Series(self.lr.coef_[i], index=[(f,'','')]))
            # Initialize x values with min and max
            xPlot[f] = [Xorig[f].min(), Xorig[f].max()]

        # Iterate over first-degree rules
        for i in range(self.z.shape[1]):
            if self.z.iloc[:,i].sum() > 1:
                continue
            # MultiIndex of rule
            idxTerm = self.z.index[self.z.iloc[:,i] > 0]
            (f, o, v) = idxTerm[0]
            # Restrict to specified features
            if (features is not None) and (f not in features):
                continue
            # Append new term
            terms = terms.append(pd.Series(self.lr.coef_[i+nnzOrd], index=idxTerm))
            # Update x values
            if f not in xPlot:
                if o in ['<=','>']:
                    # Ordinal feature, initialize with min and max
                    xPlot[f] = [Xorig[f].min(), Xorig[f].max()]
                else:
                    # Categorical feature, use all values
                    xPlot[f] = np.sort(Xorig[f].unique())
            if o in ['<=','>']:
                # Append values around threshold
                xPlot[f].extend([v-self.eps, v+self.eps])

        # Initialize y values for plots and variance calculation
        yPlot = {}
        yVar = pd.DataFrame(0., index=Xorig.index, columns=xPlot.keys())
        plotLine = pd.Series(False, index=xPlot.keys())
        # Iterate over GAM features
        for f in xPlot.keys():
            # Sort x values
            xPlot[f] = np.sort(np.array(xPlot[f]))
            yPlot[f] = np.zeros_like(xPlot[f], dtype=float)
            # Iterate over terms involving feature
            for ((o,v), c) in terms[f].iteritems():
                if o == '':
                    if self.useOrd and (f in fb.ordinal):
                        # Add linear function of standardized feature with same factor of 0.4
                        idxf = fb.ordinal.index(f)
                        yPlot[f] += 0.4 * c * (xPlot[f] - fb.scaler.mean_[idxf]) / fb.scaler.scale_[idxf]
                        yVar[f] += 0.4 * c * (Xorig[f] - fb.scaler.mean_[idxf]) / fb.scaler.scale_[idxf]
                        plotLine[f] = True
                    else:
                        # Binary feature, add indicator function
                        yPlot[f] += c * (xPlot[f] == fb.maps[f].index[1])
                        yVar[f] += c * (Xorig[f] == fb.maps[f].index[1])
                elif o == '<=':
                    # Add step function
                    yPlot[f] += c * (xPlot[f] <= v)
                    yVar[f] += c * (Xorig[f] <= v)
                    plotLine[f] = True
                elif o == '>':
                    # Add step function
                    yPlot[f] += c * (xPlot[f] > v)
                    yVar[f] += c * (Xorig[f] > v)
                    plotLine[f] = True
                elif o == '==':
                    # Add indicator function
                    yPlot[f] += c * (xPlot[f].astype(str) == v)
                    yVar[f] += c * (Xorig[f].astype(str) == v)
                elif o == '!=':
                    # Add indicator function
                    yPlot[f] += c * (xPlot[f].astype(str) != v)
                    yVar[f] += c * (Xorig[f].astype(str) != v)
                elif o == 'not':
                    # Binary feature, add indicator function
                    yPlot[f] += c * (xPlot[f] == fb.maps[f].index[0])
                    yVar[f] += c * (Xorig[f] == fb.maps[f].index[0])

        # Plot in order of descending variance
        figs = {}
        yVar2 = yVar.var().sort_values(ascending=False)
        for f in yVar2.index:
            figs[f] = plt.figure()
            ax = figs[f].add_subplot(111)
            if plotLine[f]:
                ax.plot(xPlot[f], yPlot[f])
            else:
                ax.bar(xPlot[f], yPlot[f])
                ax.xaxis.set_ticks(xPlot[f])
            plt.xlabel(f)
            plt.ylabel('contribution to prediction')
        return figs, yVar2



from __future__ import print_function

from aix360.algorithms.lwbe import LocalWBExplainer

from .CEM_aen import AEADEN

import random
import numpy as np


class CEMExplainer(LocalWBExplainer):
    """
    CEMExplainer can be used to compute contrastive explanations for image and tabular data.
    This is achieved by finding what is minimally sufficient (PP - Pertinent Positive) and
    what should be necessarily absent (PN - Pertinent Negative) to maintain the original classification.
    We use elastic norm regularization to ensure minimality for both parts of the explanation
    i.e. PPs and PNs. An autoencoder can optionally be used to make the explanations more realistic. [#]_

    References:
        .. [#] `Amit Dhurandhar, Pin-Yu Chen, Ronny Luss, Chun-Chen Tu,
           Paishun Ting, Karthikeyan Shanmugam, Payel Das, "Explanations based on
           the Missing: Towards Contrastive Explanations with Pertinent Negatives,"
           Advances in Neural Information Processing Systems (NeurIPS), 2018.
           <https://arxiv.org/abs/1802.07623>`_
    """
    def __init__(self, model):

        """
        Constructor method, initializes the explainer

        Args:
            model: KerasClassifier model whose predictions needs to be explained
        """
        super(CEMExplainer, self).__init__()
        self._wbmodel = model


    def set_params(self, *argv, **kwargs):
        """
        Set parameters for the explainer.
        """
        pass


    def explain_instance(self, input_X,
                         arg_mode, AE_model, arg_kappa, arg_b,
                         arg_max_iter, arg_init_const, arg_beta, arg_gamma):

        """
        Explains an input instance input_X and returns contrastive explanations. 
        Note that this assumes that the classifier was trained with inputs normalized in [-0.5,0.5] range.

        Args:
            input_X (numpy.ndarray): input instance to be explained
            arg_mode (str): 'PP' or 'PN'
            AE_model: Auto-encoder model
            arg_kappa (double): Confidence gap between desired class and other classes
            arg_b (double): Number of different weightings of loss function to try
            arg_max_iter (int): For each weighting of loss function number of iterations to search
            arg_init_const (double): Initial weighting of loss function
            arg_beta (double): Weighting of L1 loss
            arg_gamma (double): Weighting of auto-encoder

        Returns:
            tuple:
                * **adv_X** (`numpy ndarray`) -- Perturbed input instance for PP/PN
                * **delta_X** (`numpy ndarray`) -- Difference between input and Perturbed instance
                * **INFO** (`str`) -- Other information about PP/PN
        """

        random.seed(121)
        np.random.seed(1211)

        (_, orig_class, orig_prob_str) = self._wbmodel.predict_long(input_X)
        target_label = orig_class

        target = np.array([np.eye(self._wbmodel._nb_classes)[target_label]])

        # Hard coding batch_size=1
        batch_size = 1

        # Example: for MNIST (1, 28, 28, 1), for tabular (1, no of columns)
        shape = input_X.shape

        attack = AEADEN(self._wbmodel, shape,
                        mode=arg_mode, AE=AE_model, batch_size=batch_size,
                        kappa=arg_kappa, init_learning_rate=1e-2,
                        binary_search_steps=arg_b, max_iterations=arg_max_iter,
                        initial_const=arg_init_const, beta=arg_beta, gamma=arg_gamma)

        adv_X = attack.attack(input_X, target)

        adv_prob, adv_class, adv_prob_str = self._wbmodel.predict_long(adv_X)

        delta_X = input_X - adv_X

        _, delta_class, delta_prob_str = self._wbmodel.predict_long(delta_X)

        INFO = "[INFO]kappa:{}, Orig class:{}, Perturbed class:{}, Delta class: {}, Orig prob:{}, Perturbed prob:{}, Delta prob:{}".format(
            arg_kappa, orig_class, adv_class, delta_class, orig_prob_str, adv_prob_str, delta_prob_str)

        return (adv_X, delta_X, INFO)



## based on setup_mnist.py -- mnist data and model loading code
##
## Original Work Copyright (C) 2016, Nicholas Carlini <nicholas@carlini.com>.
##
##   This program is licenced under the BSD 2-Clause licence,
##   contained in the LICENCE file in this directory.
## Modifications Copyright (2019) IBM Corporation


import numpy as np
import os
import gzip
import urllib.request


    
def extract_data(filename, num_images):
    with gzip.open(filename) as bytestream:
        bytestream.read(16)
        buf = bytestream.read(num_images*28*28)
        data = np.frombuffer(buf, dtype=np.uint8).astype(np.float32)
        data = (data / 255) - 0.5
        data = data.reshape(num_images, 28, 28, 1)
        return data
    
def extract_labels(filename, num_images):
    with gzip.open(filename) as bytestream:
        bytestream.read(8)
        buf = bytestream.read(1 * num_images)
        labels = np.frombuffer(buf, dtype=np.uint8)
    return (np.arange(10) == labels[:, None]).astype(np.float32)
    
class MNISTDataset():
    def __init__(self, custom_preprocessing=None, dirpath=None): 
        self._dirpath = dirpath
        if not self._dirpath:
            self._dirpath = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                '..', 'data','mnist_data')

        files = ["train-images-idx3-ubyte.gz",
                 "t10k-images-idx3-ubyte.gz",
                 "train-labels-idx1-ubyte.gz",
                 "t10k-labels-idx1-ubyte.gz"]
        for name in files:
            if not os.path.exists(self._dirpath + "/" + name):
                print("retrieving file", name)
                urllib.request.urlretrieve('http://yann.lecun.com/exdb/mnist/' + name, self._dirpath + "/" + name)
                print("retrieved")

        train_data       = extract_data(self._dirpath + "/train-images-idx3-ubyte.gz", 60000)
        train_labels     = extract_labels(self._dirpath + "/train-labels-idx1-ubyte.gz", 60000)
        self.test_data   = extract_data(self._dirpath + "/t10k-images-idx3-ubyte.gz", 10000)
        self.test_labels = extract_labels(self._dirpath + "/t10k-labels-idx1-ubyte.gz", 10000)
        
        VALIDATION_SIZE = 5000
        
        self.validation_data = train_data[:VALIDATION_SIZE, :, :, :]
        self.validation_labels = train_labels[:VALIDATION_SIZE]
        self.train_data = train_data[VALIDATION_SIZE:, :, :, :]
        self.train_labels = train_labels[VALIDATION_SIZE:]
        
        



import setuptools

version = '0.1.0'

with open("aix360/version.py", 'w') as f:
    f.write('# generated by setup.py\nversion = "{}"\n'.format(version))

setuptools.setup(
	name='aix360',
	version=version,
        description='IBM AI Explainability 360',
        authos='aix360 developers',
        url='https://github.com/IBM/AIX360',
        author_email='aix360@us.ibm.com',
	packages=setuptools.find_packages(),
	license='Apache License 2.0',
	long_description=open("README.md", 'r').read(),
        install_requires=[
                'joblib>=0.11',
                'scikit-learn>=0.21.2',
                'torch',
                'torchvision',
                'cvxpy',
                'cvxopt',
                'Image',
                'keras',
                'matplotlib',
		'numpy',
                'pandas',
                'scipy>=0.17',
                'tensorflow',
                'xport',
		'scikit-image',
		'requests'
	],
        package_data={'aix360': ['data/*', 'data/*/*', 'data/*/*/*', 'models/*', 'models/*/*', 'models/*/*/*']},
	include_package_data=True,
	zip_safe=False
)



from aix360.algorithms.dise import DISExplainer
from sklearn import metrics

class TED_CartesianExplainer(DISExplainer):
    """TED is an explainability framework that leverages domain-relevant explanations in
    the training dataset to predict both labels and explanations for new instances [#]_.
    This an implementation of the simplest instantiation of TED, called the Cartesian Product.

    References:
        .. [#] `Michael Hind, Dennis Wei, Murray Campbell, Noel C. F. Codella,
           Amit Dhurandhar, Aleksandra Mojsilovic, Karthikeyan Natesan Ramamurthy,
           Kush R. Varshney, "TED: Teaching AI to Explain its Decisions,"
           AAAI /ACM Conference on Artificial Intelligence, Ethics,
           and Society (AIES-19), 2019.
           <http://www.aies-conference.com/wp-content/papers/main/AIES-19_paper_128.pdf>`_
    """

    def __init__(self, model):
        """
        Args:
            model (sklearn.base.BaseEstimator): a binary estimator for classification,
                i.e., it implements fit and predict.
        """
        super(TED_CartesianExplainer, self).__init__()
        self.model = model
        self.NumE = -1                  # set to dummy value to ensure it is computed before used

    def set_params(self, *argv, **kwargs):
        """
        Set parameters for the explainer.
        """
        pass

    def fit(self, X, Y, E) :
        """Train a classifier based on features (X), labels (Y), and explanations (E)

        Args:
            X: list of features vectors
            Y: list of labels
            E: list of explanations
        """
        self.NumE = E.max() + 1            # create E's start at 0, so we have MaxE + 1 unique values
        YE = self._composeYE(Y, E)         # create new Labels from Y + E
        self.clf = self.model.fit(X, YE)   # train classifier

    def predict_explain(self, X) :
        """Use TED-enhanced classifier to predict label (Y) and explanation (E) for passed instance

        Args:
            X (list of ints): features

        Returns:
            tuple:

                * **Y** (`int`) -- predicted label {0,1}
                * **E** (`int`) -- predicted explanation [0..MaxE]
        """
        assert (self.NumE  > 0),"Internal Error: NumE not computed!"
        YE = self.clf.predict(X)
        Y, E = self._decomposeYE(YE)      # decompose YE label into its components
        return Y, E

    def explain(self, X):
        """Use TED-enhanced classifier to provide an explanation (E) for passed instance

        Args:
            X (list of ints) : features
        Returns:
            int: predicted explanation [0..MaxE]
        """
        Y, E = self.predict_explain(X)
        return E

    def predict(self, X):
        """ Use TED-enhanced classifier to provide an prediction (Y) for passed instance

        Args:
            X (list of ints) : features

        Returns:
           int: predicted label {0,1}
        """
        Y, E = self.predict_explain(X)
        return Y

    def score(self, X_test, Y_test, E_test) :
        """ Evaluate the accuracy (Y and E) of the TED-enhanced classifier using a test dataset

        Args:
            X_test (list of lists): list of feature vectors
            Y_test (list of int)  : list of labels {0, 1}
            E_test (list of ints) : list of explanations {0, ..., NumExplanations -1}
        Returns:
            tuple:

                * **YE_accuracy** -- the accuracy of predictions when the labels (Y) and explanations (E) are treated as a combined label
                * **Y_accuracy** -- the prediction accuracy for labels (Y)
                * **E_accuracy** -- the prediction accuracy of explanations (E)
        """
        YE_predict = self.clf.predict(X_test)
        YE_test = self._composeYE(Y_test, E_test)
        YE_accuracy = metrics.accuracy_score(YE_test, YE_predict)

        Y_predict, E_predict = self._decomposeYE(YE_predict)
        Y_accuracy = metrics.accuracy_score(Y_test, Y_predict)
        E_accuracy = metrics.accuracy_score(E_test, E_predict)

        return YE_accuracy, Y_accuracy, E_accuracy



##############################
#### Supporting functions ####
##############################
    def _composeYE(self, Y, E) :
        """ Create cartesian product of Y and E

        Args:
            Y : list of labels
            E : list of explanations
        Returns:
            list of labels : We map Y and E into new dense label space, where::

                Y = 0 ==> 0 .. NumE-1
                Y = 1 ==> NumE .. 2*NumE - 1
        """
        return Y * self.NumE + E

    def _decomposeYE(self, YE) :
        """ Decompose an array of YE's into component Y and E

        Args:
            YE (array of ints) : the combined YE values created by _createYE
        Returns:
            tuple:

                * **Y** (`array of ints`) -- the Y labels
                * **E** (`array of ints`) -- the E labels
        """
        Y = YE // self.NumE   # integer division, i.e., result is an int (floor of math result)
        E = YE % self.NumE
        return Y, E



from aix360.algorithms.gwbe import GlobalWBExplainer
import numpy as np
import keras
from keras.callbacks import ModelCheckpoint
from keras.layers import Dense, Conv2D, BatchNormalization, Activation
from keras.layers import AveragePooling2D, Input, Flatten
from keras.optimizers import Adam
from keras.models import Model
from keras import backend as K
from aix360.algorithms.profwt.sample_weight_from_probes import prof_weight_compute

class ProfweightExplainer(GlobalWBExplainer):
    """
    The main explainer class that implements functions for computation of sample-wise weights from probe confidences of different layers from
    the complex model and then retrains the simple model with those weights. Implements the technique in the following reference: [#]_.

    References:
        .. [#] `Dhurandhar, Shanmugam, Luss, Olsen. Improving Simple Models with Confidence Profiles. NeurIPS 2018 <https://arxiv.org/abs/1807.07506>`_

    """
    def __init__(self):

        """
        Initialize ProfweightExplainer.
        """
        super(ProfweightExplainer, self).__init__()
        #self.model_args = model_args
        self._model = None

    def set_params(self, *argv, **kwargs):
        """
        Set parameters for the explainer.
        """
        print("TBD: Implement set params in DIPVAEExplainer")

    def explain(self,x_train,y_train,x_test,y_test,simple_model,hps,list_probe_filenames,start_layer,end_layer,model_type='neural_keras'):
        """
        Obtains (train) sample-wise weights from stored probe confidences for different layers of the complex moddel. Retrains a simple
        model using the corresponding weighted training set.

        Parameters:
            x_train (numpy array): Dataset of features to retrain the simple model using weights derived from probe confidences. Dimensions (num of samples x feature dimensions)
            y_train (numpy array): Labels for the dataset that is used to retrain simple model using weights. Dimensions (num of samples x num of classes)
            x_test (numpy array): Test dataset to evaluate the simple model trained using weights. Dimensions (num of samples x feature dimensions)
            y_test (numpy array): Test labels to evaulate the simple model trained using weights. Dimensions (num of samples x num of classes)
            hps (namedtuple): Hyperparameters (usually a named tuple with entries) to train the simple model. Please see fit function to
                find out the required set of parameters expected by the fit function.
            list_probe_filenames (list of strings): List of strings indicated path to files where different probe confidences are stored.
            start_layer (int): Index corresponding to the starting layer whose probe confidences are going to be averaged to obtain weights.
                This is an index of list_probe_filenames.
            end_layer (int): Index corresponding to the last layer whose probe confidences are going to be averaged to obtain weights. This is an index of list_probe_filenames.
            model_type (string): This specifies the type of simple model to be trained. Default is 'neural_keras'. Only this option is implemented now.
            simple_model (function object for a Keras model): This is a function object that would initialize a keras model to specify the architecture of the simple model.

        Returns:
            None

        """


        w=prof_weight_compute(list_probe_filenames,start_layer,end_layer,y_train)
        # Train the simple model with Weights
        assert hasattr(hps,'checkpoint_path'), "Checkpoint to save not mentioned"
        m=self.fit(x_train,y_train,x_test,y_test,simple_model,hps,'neural_keras',w.reshape(w.shape[0],))
        print("Accuracy of Prof-Weighted Simple Model:",m[1])
        self._model=m[0]
        return


    def fit(self,x_train,y_train,x_test,y_test,simple_model,hps,model_type='neural_keras',sample_weight=None):
        """
        Fits the training data by initializing a simple model with hyper parameters and returns the test accuracy on the test dataset.
        This can be trained with or without sample weights. The first 500 samples of the test dataset is used as validation data.

        Parameters:
            x_train (numpy array): Training dataset features for training the simple model. Dimensions (num of samples x feature dimensions)
            y_train (numpy array): Labels for the training dataset to train on. Dimensions (num of samples x num of classes)
            x_test (numpy array): Test dataset features. Dimensions (num of samples x feature dimensions)
            y_test (numpy array): Test dataset labels. Dimensions (num of samples x num of classes)
            hps (namedtuple): A namedtuple that is expected to have the following named tuple elements:

                * optimizer - specified the optimizer in keras.
                * complexity_param - Used for Resenet based simple model to specify number of Resunits. Used by simple model function
                  object to intialize a simple model of appropriate complexity.
                * num_classes - scalar specifying number of classes used by the simple
                  model function.
                * checkpoint_path - specifies the path for saving a checkpoint of the trained model. This is expected.
                * lr_scheduler - a function object that takes in a scalar (epochs) and specified a learning rate (scalar). This is a learning rate Scheduler. Expected.
                * lr_reducer  - a function object that specifies how learning rates must be reduced if validation accuracy does not improve - Optional.

            simple_model (function object for a Keras model): A function object that constructs a keras model for the simple model and returns the model object. It is expected to take in input_shape, hps.complexity_param and num_classes.
                It is expected to implement a keras model fit function. It is also expected to implement a keras model evaulate function.

        Returns:
            tuple:

                * **model_d** (`Keras model object`) -- Returns the trained model that is initialized by simple_model functions.
                * **scores[1]** (`float`) -- Returns the test accuracy of the trained model on (x_test,y_test.)

        """


        input_shape=x_train.shape[1:]
        if model_type=='neural_keras':
            assert None not in (hps.complexity_param,hps.num_classes), "Missing Hyper Parameters"
            model_d=simple_model(input_shape,hps.complexity_param,hps.num_classes)
            assert hps.optimizer is not None, "Missing Optimizer Specs"
            model_d.compile(loss='categorical_crossentropy',optimizer=hps.optimizer,metrics=['accuracy'])
            if (hasattr(model_d,'summary')):
                model_d.summary()
            callbacks=[]
            assert hps.checkpoint_path is not None, "Checkpoint to save Model not specified"
            checkpoint = ModelCheckpoint(filepath=hps.checkpoint_path,monitor='val_acc',verbose=1,save_best_only=True)
            assert hps.lr_scheduler is not None, "Learning Rate Scheduler not specified"
            callbacks=callbacks+[checkpoint,hps.lr_scheduler]
            if hasattr(hps,'lr_reducer'):
                callbacks=callbacks+[hps.lr_reducer]
            assert hasattr(model_d,'fit'), "Model supplied needs a fit function"
            if sample_weight is not None:
                model_d.fit(x_train, y_train,batch_size=hps.batch_size,epochs=hps.epochs,validation_data=(x_test[0:500,:], y_test[0:500,:]),shuffle=True,callbacks=callbacks,sample_weight=sample_weight)
            else:
                model_d.fit(x_train, y_train,batch_size=hps.batch_size,epochs=hps.epochs,validation_data=(x_test[0:500,:], y_test[0:500,:]),shuffle=True,callbacks=callbacks)
            assert hasattr(model_d,'evaluate'), "Model supplied needs an evaluate function"
            scores=model_d.evaluate(x_test, y_test, verbose=1)

        return (model_d,scores[1])



from .features import FeatureBinarizer
from .linear_regression import LinearRuleRegression
from .logistic_regression import LogisticRuleRegression
from .boolean_rule_cg import BooleanRuleCG
from .GLRM import GLRMExplainer
from .BRCG import BRCGExplainer



import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
from torch.nn import Linear

import os

td = torch.distributions


def convert_and_reshape(x, dataset_obj):
    return x.cpu().data.numpy().reshape([-1] + dataset_obj.data_dims).transpose(0,2,3,1)


def plot_reconstructions(dataset_obj, trained_net, input_images, image_id_to_plot=0, epoch=0, batch_id = 0, save_dir="results"):
    z_sample, _, _ = trained_net.encode(input_images)
    #print(z_sample[image_id_to_plot])
    recons = trained_net.decode(z_sample)

    input_images_numpy = convert_and_reshape(input_images, dataset_obj)
    recons_numpy = convert_and_reshape(recons, dataset_obj)

    f, axarr = plt.subplots(1, 2)
    if input_images_numpy.shape[2] == 1:
        axarr[0].imshow(input_images_numpy[image_id_to_plot, :, :, 0], cmap="gray")
        axarr[1].imshow(recons_numpy[image_id_to_plot, :, :, 0], cmap="gray")
    else:
        axarr[0].imshow(input_images_numpy[image_id_to_plot]*0.5 + 0.5)
        axarr[1].imshow(recons_numpy[image_id_to_plot]*0.5 + 0.5)

    if not os.path.isdir(save_dir):
        os.makedirs(save_dir)

    f.savefig(os.path.join(save_dir, 'recons_epoch_{}_batch_id_{}.png'.format(epoch, batch_id)))
    plt.close(fig=f)


def plot_latent_traversal(explainer, input_images, args, dataset_obj, image_id_to_plot=0, num_sweeps=15,
                          max_abs_edit_value=10.0, epoch=0, batch_id = 0, save_dir="results"):
    edit_dim_values = np.linspace(-1.0 *max_abs_edit_value, max_abs_edit_value, num_sweeps)

    f, axarr = plt.subplots(args.latent_dim, len(edit_dim_values), sharex=True, sharey=True)
    f.set_size_inches(10, 10* args.latent_dim / len(edit_dim_values))

    for i in range(args.latent_dim):
        for j in range(len(edit_dim_values)):

            edited_images = convert_and_reshape(explainer.explain(input_images=input_images,
                             edit_dim_id = i,
                             edit_dim_value = edit_dim_values[j],edit_z_sample=False), dataset_obj)
            if edited_images.shape[2] == 1:
                axarr[i][j].imshow(edited_images[image_id_to_plot,:,:,0], cmap="gray", aspect='auto')
            else:
                axarr[i][j].imshow(edited_images[image_id_to_plot]*0.5 + 0.5, aspect='auto')
            #axarr[j][i].axis('off')
            if i == len(axarr) - 1:
                axarr[i][j].set_xlabel("z:" + str(np.round(edit_dim_values[j], 1)))
            if j == 0:
                axarr[i][j].set_ylabel("l:" + str(i))
            axarr[i][j].set_yticks([])
            axarr[i][j].set_xticks([])
    plt.subplots_adjust(hspace=0, wspace=0)

    if not os.path.isdir(save_dir):
        os.makedirs(save_dir)

    f.savefig(os.path.join(save_dir, 'traversal_epoch_{}_batch_id_{}.png'.format(epoch, batch_id)))
    plt.close(fig=f)


def reparam(mu, std, do_sample=True, cuda=True):
    """Reparametrization for Normal distribution.
    """
    if do_sample:
        eps = torch.FloatTensor(std.size()).normal_()
        if cuda:
            eps = eps.cuda()
        eps = Variable(eps)
        return mu + eps * std
    else:
        return mu


def bernoulli_likelihood(x, x_pred, dim=None):
    """Compute Bernoulli likelihood.
    """
    assert x_pred.min() >= 0.
    assert x_pred.max() <= 1.
    lik = td.Bernoulli(probs=x_pred).log_prob(x)
    if torch.isnan(lik.sum()):
        print (lik.sum())
    if dim:
        return lik.sum(dim=dim)
    else:
        return lik.sum()


def gaussian_likelihood(x, x_pred):
    """Compute Gaussian likelihood.
    """
    return - torch.sum(F.mse_loss(x_pred, x, reduction='none'))


class FCNet(nn.Module):
    def __init__(self, num_nodes=50, ip_dim=1, op_dim=1, activation_type='relu', args=None):
        super(FCNet, self).__init__()
        self.args = args
        if activation_type == 'relu':
            self.activation = nn.ReLU()
        elif activation_type == 'tanh':
            self.activation = nn.Tanh()
        else:
            print("Activation Type not supported")
            return
        layer = Linear
        self.fc_hidden = []
        self.fc1 = layer(ip_dim, num_nodes)
        self.bn1 = nn.BatchNorm1d(num_nodes)
        for _ in np.arange(self.args.num_layers - 1):
            self.fc_hidden.append(layer(num_nodes, num_nodes))
            self.fc_hidden.append(nn.BatchNorm1d(num_nodes))
            self.fc_hidden.append(self.activation)
        self.features = nn.Sequential(*self.fc_hidden)
        self.fc_out = layer(num_nodes, op_dim)

    def forward(self, x):
        x = x.squeeze().view(x.size(0), -1)
        x = self.fc1(x)
        x = self.bn1(x)
        x = self.activation(x)
        x = self.features(x)
        return self.fc_out(x)


class VAE(nn.Module):
    """
    Variational Autoencoder [1]_ is an generative model with stochastic encoder and decoder learned using Bayesian
    variational inference. The encoder network is used to obtain latent representations for the input while the decoder
    generated the samples from the latents.

    References:
        .. [1] Diederik P Kingma and Max Welling. Auto-encoding variational Bayes. ICLR, 2014
    """
    def __init__(self, num_nodes=50, latent_dim=10, op_dim=784, activation_type='relu',
                 args=None, cuda_available=None, init_networks=True):
        super(VAE, self).__init__()
        self.latent_dim = latent_dim

        if init_networks:
            args.num_layers = args.num_gen_layers
            generative_op_dim = op_dim
            self.generative_net = FCNet(num_nodes=num_nodes, op_dim=generative_op_dim, ip_dim=latent_dim,
                                        activation_type=activation_type, args=args)
            args.num_layers = args.num_inference_layers
            self.inference_net = FCNet(num_nodes=num_nodes, op_dim=2 * latent_dim, ip_dim=op_dim,
                                       activation_type=activation_type,
                                       args=args)
        self.z_prior_stdv = torch.Tensor([1.])
        self.z_prior_mean = torch.Tensor([0.])
        if cuda_available:
            self.z_prior_stdv = self.z_prior_stdv.cuda()
            self.z_prior_mean = self.z_prior_mean.cuda()
        self.cuda_available = cuda_available
        self.name = "VAE"
        self.args = args

    def encode(self, x):
        z = self.inference_net.forward(x)
        mu_z = z[:, :self.latent_dim]
        std_z = F.softplus(z[:, self.latent_dim:])
        z_sample = reparam(mu_z, std_z, cuda=self.cuda_available)
        return z_sample, mu_z, std_z

    def decode(self, z):
        return torch.sigmoid(self.generative_net(z))

    def forward(self, x, return_z_sample=False):
        z_sample, mu_z, std_z = self.encode(x)
        if return_z_sample:
            return self.generative_net(z_sample), mu_z, std_z, z_sample
        else:
            return self.generative_net(z_sample), mu_z, std_z

    def sample_from_gen_model(self, num_samples=25):
        self.eval()
        z_samples = torch.randn(num_samples, self.latent_dim)
        if self.cuda_available:
            z_samples = z_samples.cuda()
        self.train()
        return torch.sigmoid(self.generative_net(z_samples))

    def _kl_divergence_z(self, mu_z, std_z):
        """
        """
        kld_z = self.z_prior_stdv.log() - std_z.log() + (std_z ** 2 + (mu_z.pow(2) - self.z_prior_mean)) /\
                (2 * self.z_prior_stdv.pow(2)) - 0.5
        return kld_z.sum()

    def neg_elbo(self, x, beta=None):
        recon, mu_z, std_z = self.forward(x)
        recon = torch.sigmoid(recon)
        kl_z = self._kl_divergence_z(mu_z, std_z)
        if beta is not None:
            kl_z = beta * kl_z
        return - bernoulli_likelihood(x, recon) + kl_z


class DIPVAE(nn.Module):
    """
    Disentangled Inferred Prior-VAE or DIPVAE [1] is a Variational Autoencoder [2]_ variant that leads to a
    disentangled latent space. This is achieved by matching the covariance of the prior distributions with the
    inferred prior.

    References:
        .. [1] Variational Inference of Disentangled Latent Concepts from Unlabeled Observations (DIP-VAE), ICLR 2018.
         Kumar, Sattigeri, Balakrishnan. https://arxiv.org/abs/1711.00848
        .. [2] Diederik P Kingma and Max Welling. Auto-encoding variational Bayes. ICLR, 2014
    """
    def __init__(self, num_nodes=50, latent_dim=10, op_dim=784, activation_type='relu',
                 args=None, cuda_available=None, init_networks=True, mode=None, output_activation_type=None,
                 likelihood_type=None, beta=1.0):
        super(DIPVAE, self).__init__()
        self.latent_dim = latent_dim
        self.mode = mode

        if init_networks:
            args.num_layers = args.num_gen_layers
            generative_op_dim = op_dim
            self.generative_net = FCNet(num_nodes=num_nodes, op_dim=generative_op_dim, ip_dim=latent_dim,
                                        activation_type=activation_type, args=args)
            args.num_layers = args.num_inference_layers
            self.inference_net = FCNet(num_nodes=num_nodes, op_dim=2 * latent_dim, ip_dim=op_dim,
                                       activation_type=activation_type,
                                       args=args)
        self.z_prior_stdv = torch.Tensor([1.])
        self.z_prior_mean = torch.Tensor([0.])
        if cuda_available:
            self.z_prior_stdv = self.z_prior_stdv.cuda()
            self.z_prior_mean = self.z_prior_mean.cuda()
        self.cuda_available = cuda_available
        self.name = "DIPVAE"
        self.args = args

        self.beta = beta

        self.lambda_diag_factor = args.lambda_diag_factor
        self.lambda_offdiag = args.lambda_offdiag
        self.lambda_diag = self.lambda_diag_factor * self.lambda_offdiag

        self.output_activation_type = output_activation_type
        if likelihood_type == "bernoulli":
            self.likelihood = bernoulli_likelihood
        elif likelihood_type == "gaussian":
            self.likelihood = gaussian_likelihood
        else:
            raise NotImplementedError("Unsupported likelihood type.")

    def encode(self, x):
        z = self.inference_net.forward(x)
        mu_z = z[:, :self.latent_dim]
        std_z = F.softplus(z[:, self.latent_dim:])
        z_sample = reparam(mu_z, std_z, cuda=self.cuda_available)
        return z_sample, mu_z, std_z

    def decode(self, z):
        if self.output_activation_type is None:
            return self.generative_net(z)
        elif self.output_activation_type == "sigmoid":
            return torch.sigmoid(self.generative_net(z))
        elif self.output_activation_type == "tanh":
            return torch.tanh(self.generative_net(z))
        else:
            raise NotImplementedError("Unsupported output activation type.")

    def forward(self, x, return_z_sample=False):
        z_sample, mu_z, std_z = self.encode(x)
        if return_z_sample:
            return self.decode(z_sample), mu_z, std_z, z_sample
        else:
            return self.decode(z_sample), mu_z, std_z

    def sample_from_gen_model(self, num_samples=25):
        self.eval()
        z_samples = torch.randn(num_samples, self.latent_dim)
        if self.cuda_available:
            z_samples = z_samples.cuda()
        self.train()
        return self.decode(z_samples)

    def _get_covariance_mu_z(self, mu_z):

        zero_mean_mu_z = mu_z - torch.mean(mu_z, dim=1, keepdim=True)
        zero_mean_mu_z_t = zero_mean_mu_z.t()

        return zero_mean_mu_z_t.matmul(zero_mean_mu_z).squeeze()

    def _get_dipvae_regularizer(self, cov_z, lambda_offdiag, lambda_diag):
        cov_z_diag = torch.diag(cov_z)
        cov_z_offdiag = cov_z - torch.diag(cov_z_diag)

        return lambda_offdiag * torch.sum(cov_z_offdiag ** 2) + lambda_diag * torch.sum((cov_z_diag - 1) ** 2)

    def _regularizer(self, mu_z, std_z):
        kld_z = self.z_prior_stdv.log() - std_z.log() + (std_z ** 2 + (mu_z.pow(2) - self.z_prior_mean)) /\
                (2 * self.z_prior_stdv.pow(2)) - 0.5
        regularizer_loss = kld_z.sum()

        regularizer_loss = self.beta * regularizer_loss

        cov_mu_z = self._get_covariance_mu_z(mu_z)

        if self.mode == "i":
            dipvae_regularizer_loss = self._get_dipvae_regularizer(cov_mu_z, self.lambda_offdiag, self.lambda_diag)

        elif self.mode == "ii":
            cov_z = cov_mu_z + torch.mean(torch.diag(std_z**2), dim=0)
            dipvae_regularizer_loss = self._get_dipvae_regularizer(cov_z, self.lambda_offdiag, self.lambda_diag)
        else:
            raise NotImplementedError("Unsupported dipvae mode.")

        return regularizer_loss + dipvae_regularizer_loss

    def neg_elbo(self, x):
        recon, mu_z, std_z = self.forward(x)
        regularizer = self._regularizer(mu_z, std_z)
        return - self.likelihood(x, recon) + regularizer
